@@ -11,11 +11,6 @@
 #include "func.h"
 #include "kanji.h"
 
-#if	!MSDOS
-#include <sys/file.h>
-#include <sys/param.h>
-#endif
-
 typedef struct _attr_t {
 	u_short mode;
 #ifdef	HAVEFLAGS
@@ -129,11 +124,17 @@ char *path, *org;
 static int NEAR islowerdir(path, org)
 char *path, *org;
 {
+#if	!MSDOS && !defined (_NODOSDRIVE)
+	char orgpath[MAXPATHLEN];
+#endif
 	char *cp, *top, *cwd, tmp[MAXPATHLEN];
 	int i;
 
 	cwd = getwd2();
 	if (!org) org = cwd;
+#if	!MSDOS && !defined (_NODOSDRIVE)
+	else org = nodospath(orgpath, org);
+#endif
 	strcpy(tmp, path);
 	top = tmp;
 #if	MSDOS || !defined (_NODOSDRIVE)
@@ -333,7 +334,7 @@ struct stat *stp1, *stp2;
 			case 2:
 				if ((cp = strrdelim(dest, 1))) cp++;
 				else cp = dest;
-				lcmdline = LINFO - n_line;
+				lcmdline = LINFO;
 				if (!(tmp = inputstr(NEWNM_K, 1,
 					-1, NULL, -1))) return(-1);
 				strcpy(cp, tmp);
@@ -406,7 +407,7 @@ int mode;
 	{
 		char *tmp, dir[MAXPATHLEN];
 
-		if (_Xlstat(path, &st) < 0) {
+		if (_Xlstat(path, &st, 0, 1) < 0) {
 			warning(-1, path);
 			return(-1);
 		}
@@ -416,7 +417,7 @@ int mode;
 		else if (tmp == path) strcpy(dir, _SS_);
 		else strncpy2(dir, path, tmp - path);
 
-		if (_Xlstat(dir, &st) < 0) {
+		if (_Xlstat(dir, &st, 0, 1) < 0) {
 			warning(-1, dir);
 			return(-1);
 		}
@@ -520,7 +521,7 @@ char *path;
 			break;
 	}
 	if (destdir) free(destdir);
-	destdir = strdup2(dest + strlen(destpath) + 1);
+	destdir = strdup2(&(dest[strlen(destpath) + 1]));
 	return(0);
 }
 
@@ -647,10 +648,10 @@ int y;
 	kanjiputs(TMODE_K);
 	locate(n_column / 2, y);
 	putmode(buf, listp -> st_mode);
-	cputs2(buf + 1);
+	cputs2(&(buf[1]));
 	locate(n_column / 2 + 10, y);
 	putmode(buf, attr -> mode);
-	cputs2(buf + 1);
+	cputs2(&(buf[1]));
 
 #ifdef	HAVEFLAGS
 	locate(0, ++y);
@@ -691,7 +692,7 @@ int y;
 
 int inputattr(listp, flag)
 namelist *listp;
-u_short flag;
+int flag;
 {
 	struct tm *tm;
 	attr_t attr;
@@ -852,7 +853,7 @@ u_short flag;
 #endif
 				locate(n_column / 2 + 10, yy + y + 2);
 				putmode(buf, attr.mode);
-				cputs2(buf + 1);
+				cputs2(&(buf[1]));
 				break;
 			default:
 				break;
@@ -923,10 +924,21 @@ int applyfile(func, endmes)
 int (*func)__P_((char *));
 char *endmes;
 {
+#if	!MSDOS && !defined (_NODOSDRIVE)
+	char path[MAXPATHLEN];
+#endif
 	int i, ret, old, dupfilepos;
 
 	dupfilepos = filepos;
 	ret = old = filepos;
+
+	if (mark <= 0) {
+		i = (*func)(fnodospath(path, filepos));
+		if (i < 0) warning(-1, filelist[filepos].name);
+		else if (!i) ret++;
+		return(ret);
+	}
+
 	helpbar();
 	for (filepos = 0; filepos < maxfile; filepos++) {
 		if (intrkey()) {
@@ -938,9 +950,9 @@ char *endmes;
 		movepos(old, 0);
 		locate(win_x, win_y);
 		tflush();
-		if ((i = (*func)(filelist[filepos].name)) < 0)
-			warning(-1, filelist[filepos].name);
-		if (!i && ret == filepos) ret++;
+		i = (*func)(fnodospath(path, filepos));
+		if (i < 0) warning(-1, filelist[filepos].name);
+		else if (!i && ret == filepos) ret++;
 
 		old = filepos;
 	}
@@ -983,6 +995,7 @@ int *maxp, depth;
 			}
 		}
 	}
+	if (cp > dir) *(cp - 1) = '\0';
 	Xclosedir(dirp);
 	return(dirlist);
 }
@@ -1013,10 +1026,11 @@ char *endmes;
 	tflush();
 
 	if (!funcd1 && !funcd2) order = 0;
-	fname = strcatdelim2(path, dir, NULL);
+	strcpy(path, dir);
 
 	ret = max = 0;
 	dirlist = getdirtree(path, NULL, &max, (order == 2) ? 0 : 1);
+	fname = strcatdelim(path);
 
 	destmtime = destatime = (time_t)-1;
 	if ((order == 1 || order == 2) && (ret = (*funcd1)(dir)) < 0) {
@@ -1093,6 +1107,9 @@ char *endmes;
 		realpath2(dir, path, 0);
 		dir = path;
 	}
+#if	!MSDOS && !defined (_NODOSDRIVE)
+	else dir = nodospath(path, dir);
+#endif
 	return(_applydir(dir, funcf, funcd1, funcd2, order, endmes));
 }
 
@@ -1121,9 +1138,10 @@ int tr;
 #ifndef	_NODOSDRIVE
 	if (dospath3("")) waitmes();
 #endif
-	if (mark > 0) applyfile(safecopy, ENDCP_K);
-	else if (isdir(&(filelist[filepos]))
-	&& !islink(&(filelist[filepos]))) {
+	if (mark > 0
+	|| (!isdir(&(filelist[filepos])) || islink(&(filelist[filepos]))))
+		applyfile(safecopy, ENDCP_K);
+	else {
 #ifdef	_NOEXTRACOPY
 		if (copypolicy) {
 			warning(EEXIST, filelist[filepos].name);
@@ -1135,8 +1153,7 @@ int tr;
 			(islowerdir(destpath, filelist[filepos].name)) ? 2 : 1,
 			ENDCP_K);
 	}
-	else if (safecopy(filelist[filepos].name) < 0)
-		warning(-1, filelist[filepos].name);
+
 	free(destpath);
 	if (destdir) free(destdir);
 #ifndef	_NOEXTRACOPY
@@ -1180,21 +1197,30 @@ int tr;
 	if (mark > 0) filepos = applyfile(safemove, ENDMV_K);
 	else if (islowerdir(destpath, filelist[filepos].name))
 		warning(EINVAL, filelist[filepos].name);
-	else if ((i = safemove(filelist[filepos].name)) < 0) {
-#ifdef	_NOEXTRACOPY
-		warning(-1, filelist[filepos].name);
-#else
-		if (errno != EXDEV
-		|| !isdir(&(filelist[filepos]))
-		|| islink(&(filelist[filepos])))
-			warning(-1, filelist[filepos].name);
-		else if (applydir(filelist[filepos].name,
-			safemove, cpdir, mvdir,
-			(islowerdir(destpath, filelist[filepos].name)) ? 2 : 1,
-			ENDMV_K) >= 0) filepos++;
+	else {
+#if	!MSDOS && !defined (_NODOSDRIVE)
+		char path[MAXPATHLEN];
 #endif
+
+		i = safemove(fnodospath(path, filepos));
+		if (!i) filepos++;
+		else if (i < 0) {
+#ifdef	_NOEXTRACOPY
+			warning(-1, filelist[filepos].name);
+#else
+			if (errno != EXDEV
+			|| !isdir(&(filelist[filepos]))
+			|| islink(&(filelist[filepos])))
+				warning(-1, filelist[filepos].name);
+			else if (applydir(filelist[filepos].name,
+				safemove, cpdir, mvdir,
+				(islowerdir(destpath, filelist[filepos].name))
+				? 2 : 1,
+				ENDMV_K) >= 0) filepos++;
+#endif
+		}
 	}
-	else if (!i) filepos++;
+
 	if (filepos >= maxfile) filepos = maxfile - 1;
 	free(destpath);
 	if (destdir) free(destdir);
