@@ -15,11 +15,6 @@
 #include "system.h"
 #endif
 
-#if	!MSDOS
-#include <sys/file.h>
-#include <sys/param.h>
-#endif
-
 #ifndef	_NODOSDRIVE
 extern int flushdrv __P_((int, VOID_T (*)__P_((VOID_A))));
 #endif
@@ -42,7 +37,7 @@ extern int win_y;
 extern char *destpath;
 extern int savehist;
 extern int sizeinfo;
-#ifndef	_NOSPLITWIN
+#if	!defined (_NOARCHIVE) || !defined (_NOSPLITWIN)
 extern char fullpath[];
 #endif
 #ifndef	_NOORIGSHELL
@@ -279,10 +274,10 @@ bindtable bindlist[MAXBINDTABLE] = {
 	{K_BEG,		CUR_TOP,	255},
 	{K_EOL,		CUR_BOTTOM,	255},
 	{K_HELP,	HELP_MESSAGE,	255},
-	{CTRL('@'),	MARK_FILE3,	255},
-	{CTRL('L'),	REREAD_DIR,	255},
-	{CTRL('R'),	SEARCH_BACK,	255},
-	{CTRL('S'),	SEARCH_FORW,	255},
+	{K_CTRL('@'),	MARK_FILE3,	255},
+	{K_CTRL('L'),	REREAD_DIR,	255},
+	{K_CTRL('R'),	SEARCH_BACK,	255},
+	{K_CTRL('S'),	SEARCH_FORW,	255},
 	{-1,		NO_OPERATION,	255}
 };
 
@@ -325,14 +320,14 @@ char *arg;
 	int n;
 
 	if (!arg || (n = atoi2(arg)) <= 0) n = 1;
-	if (filepos + FILEPERLOW * n < maxfile) filepos += FILEPERLOW * n;
+	if (filepos + FILEPERROW * n < maxfile) filepos += FILEPERROW * n;
 	else if (filepos / FILEPERPAGE == (maxfile - 1) / FILEPERPAGE)
 		return(0);
 	else {
-		filepos = ((maxfile - 1) / FILEPERLOW) * FILEPERLOW
-			+ filepos % FILEPERLOW;
+		filepos = ((maxfile - 1) / FILEPERROW) * FILEPERROW
+			+ filepos % FILEPERROW;
 		if (filepos >= maxfile) {
-			filepos -= FILEPERLOW;
+			filepos -= FILEPERROW;
 			if (filepos / FILEPERPAGE
 			< (maxfile - 1) / FILEPERPAGE)
 				filepos = maxfile - 1;
@@ -346,10 +341,10 @@ char *arg;
 {
 	int n;
 
-	if (filepos < FILEPERLOW) return(0);
+	if (filepos < FILEPERROW) return(0);
 	if (!arg || (n = atoi2(arg)) <= 0) n = 1;
-	if (filepos - FILEPERLOW * n >= 0) filepos -= FILEPERLOW * n;
-	else filepos = filepos % FILEPERLOW;
+	if (filepos - FILEPERROW * n >= 0) filepos -= FILEPERROW * n;
+	else filepos = filepos % FILEPERROW;
 	return(2);
 }
 
@@ -405,8 +400,31 @@ char *arg;
 {
 	int i, m;
 
-	if (islink(&(filelist[filepos]))) i = 0;
-	else {
+#if	!MSDOS
+	if (islink(&(filelist[filepos]))) {
+# ifndef	_NODOSDRIVE
+		char path[MAXPATHLEN];
+# endif
+		char tmp[MAXPATHLEN];
+
+# ifndef	_NOARCHIVE
+		if (archivefile) {
+			if (!(filelist[filepos].linkname)) i = 0;
+			else {
+				i = strlen(filelist[filepos].linkname);
+				if (i > sizeof(tmp) - 1) i = sizeof(tmp) - 1;
+				strncpy(tmp, filelist[filepos].linkname, i);
+			}
+		}
+		else
+# endif
+		i = Xreadlink(fnodospath(path, filepos), tmp, sizeof(tmp) - 1);
+		tmp[i] = '\0';
+		i = 1 - (strlen3(tmp) + 4);
+	}
+	else
+#endif	/* !MSDOS */
+	{
 		i = calcwidth();
 		if (isdisptyp(dispmode)) {
 			m = filelist[filepos].st_mode;
@@ -455,11 +473,13 @@ char *arg;
 static VOID NEAR markcount(VOID_A)
 {
 	char buf[14 + 1];
+	int x;
 
-	locate(CMARK + 5, LSTATUS);
+	x = (isleftshift()) ? 1 : 0;
+	locate(C_MARK + 5 - x, L_STATUS);
 	cputs2(ascnumeric(buf, mark, 4, 4));
 	if (sizeinfo) {
-		locate(CSIZE + 5, LSIZE);
+		locate(C_SIZE + 5 - x, L_SIZE);
 		cputs2(ascnumeric(buf, marksize, 3, 14));
 	}
 }
@@ -590,7 +610,8 @@ char *arg;
 	if (archivefile);
 	else
 #endif
-	if (!strcmp(filelist[filepos].name, ".")) return(warning_bell(arg));
+	if (!isdir(&(filelist[filepos]))
+	|| !strcmp(filelist[filepos].name, ".")) return(warning_bell(arg));
 	return(5);
 }
 
@@ -676,6 +697,9 @@ char *arg;
 static int log_dir(arg)
 char *arg;
 {
+#ifndef	_NOARCHIVE
+	char *cp, dupfullpath[MAXPATHLEN];
+#endif
 	char *path;
 
 	if (arg && *arg) path = strdup2(arg);
@@ -684,12 +708,32 @@ char *arg;
 		free(path);
 		return(1);
 	}
+#ifndef	_NOARCHIVE
+	if (archivefile && *path != '/') {
+		if (!(cp = archchdir(path))) {
+			warning(-1, path);
+			free(path);
+			return(1);
+		}
+		free(path);
+		filelist[filepos].name = cp;
+		return(4);
+	}
+	else
+#endif
 	if (chdir3(path) < 0) {
 		warning(-1, path);
 		free(path);
 		return(1);
 	}
 	free(path);
+#ifndef	_NOARCHIVE
+	if (archivefile) {
+		strcpy(dupfullpath, fullpath);
+		while (archivefile) poparchdupl();
+		strcpy(fullpath, dupfullpath);
+	}
+#endif
 	replacefname(NULL);
 	return(4);
 }
@@ -804,7 +848,7 @@ static int NEAR execshell(VOID_A)
 
 	n = sigvecset(0);
 	if (n) putterms(t_end);
-	stdiomode();
+	stdiomode(0);
 	kanjifputs(SHEXT_K, stderr);
 	fputc('\n', stderr);
 #ifdef	_NOORIGSHELL
@@ -814,7 +858,7 @@ static int NEAR execshell(VOID_A)
 	else ret = shell_loop(1);
 #endif
 	if (n) putterms(t_init);
-	ttyiomode();
+	ttyiomode(0);
 	sigvecset(n);
 
 	return(ret);
@@ -837,14 +881,15 @@ char *arg;
 #ifndef	_NODOSDRIVE
 	drive = 0;
 #endif
-	if (isdir(&(filelist[filepos]))
+	if (isdir(&(filelist[filepos]))) return(1);
 #ifndef	_NOARCHIVE
-	|| (archivefile && !(dir = tmpunpack(1)))
+	else if (archivefile) {
+		if (!(dir = tmpunpack(1))) return(1);
+	}
 #endif
 #ifndef	_NODOSDRIVE
-	|| (drive = tmpdosdupl("", &dir, 1)) < 0
+	else if ((drive = tmpdosdupl("", &dir, 1)) < 0) return(1);
 #endif
-	) return(1);
 
 	if (!execenv("FD_PAGER", filelist[filepos].name)) {
 #ifdef	PAGER
@@ -852,9 +897,10 @@ char *arg;
 #else
 		do {
 			dump(filelist[filepos].name);
-		} while(!yesno(PEND_K));
+		} while (!yesno(PEND_K));
 #endif
 	}
+
 #ifndef	_NODOSDRIVE
 	if (drive) removetmp(dir, NULL, filelist[filepos].name);
 	else
@@ -992,7 +1038,7 @@ char *arg;
 	int drive;
 #endif
 
-	getwsize(80, WHEADERMAX + WFOOTER + WFILEMIN);
+	checkscreen(WCOLUMNMIN, WHEADERMAX + WFOOTER + WFILEMIN);
 #ifndef	_NODOSDRIVE
 	if ((drive = dospath3(""))) flushdrv(drive, NULL);
 #endif
@@ -1004,6 +1050,10 @@ char *arg;
 static int edit_config(arg)
 char *arg;
 {
+	if (FILEPERROW < WFILEMINCUSTOM) {
+		warning(0, NOROW_K);
+		return(1);
+	}
 	customize();
 	return(4);
 }
@@ -1039,8 +1089,12 @@ char *arg;
 	char *path;
 
 	if (arg && *arg) path = strdup2(arg);
-	else if (!(path = inputstr(MAKED_K, 1, -1, NULL, 1))
-	|| !*(path = evalpath(path, 1))) return(1);
+	else if (!(path = inputstr(MAKED_K, 1, -1, NULL, 1))) return(1);
+	else if (!*(path = evalpath(path, 1))) {
+		free(path);
+		return(1);
+	}
+
 	if (mkdir2(path, 0777) < 0) warning(-1, path);
 	free(path);
 	return(4);
@@ -1095,10 +1149,8 @@ char *arg;
 		}
 	}
 	else for (;;) {
-		if (!(file =
-		inputstr(NEWNM_K, 1, 0, filelist[filepos].name, -1))
-		|| !*(file = evalpath(file, 1)))
-			return(1);
+		file = inputstr(NEWNM_K, 1, 0, filelist[filepos].name, -1);
+		if (!file || !*file) return(1);
 		if (Xaccess(file, F_OK) < 0) {
 			if (errno == ENOENT) break;
 			warning(-1, file);
@@ -1107,12 +1159,7 @@ char *arg;
 		free(file);
 	}
 
-#if	MSDOS || defined (_NODOSDRIVE)
-	if (Xrename(filelist[filepos].name, file) < 0)
-#else
-	if (Xrename(nodospath(path, filelist[filepos].name), file) < 0)
-#endif
-	{
+	if (Xrename(fnodospath(path, filepos), file) < 0) {
 		warning(-1, file);
 		free(file);
 		return(1);
@@ -1151,15 +1198,12 @@ char *arg;
 	removepolicy = 0;
 #if	!MSDOS
 	if (islink(&(filelist[filepos]))) {
+# ifndef	_NODOSDRIVE
+		char path[MAXPATHLEN];
+# endif
 		int ret;
 
-# ifdef	_NODOSDRIVE
-		ret = rmvfile(filelist[filepos].name);
-# else
-		char path[MAXPATHLEN];
-
-		ret = rmvfile(nodospath(path, filelist[filepos].name));
-# endif
+		ret = rmvfile(fnodospath(path, filepos));
 		if (ret < 0) warning(-1, filelist[filepos].name);
 		else if (!ret) filepos++;
 	}
@@ -1227,9 +1271,14 @@ char *arg;
 	int ret;
 
 	if (arg) com = strdup2(arg);
-	else if (!(com = inputshellloop(-1, NULL))) return(1);
+	else {
+		ttyiomode(1);
+		com = inputshellloop(-1, NULL);
+		ttyiomode(0);
+		if (!com) return(1);
+	}
 	if (*com) {
-		ret = execusercomm(com, filelist[filepos].name, 0, 1, 0);
+		ret = execusercomm(com, filelist[filepos].name, 0, -1, 0);
 		ret = (ret < 0) ? 1 :
 			(internal_status >= -1) ? internal_status: 4;
 	}
@@ -1266,7 +1315,9 @@ char *arg;
 #if	MSDOS && !defined (DISMISS_CURPATH)
 	len = (!isdir(&(filelist[filepos])) && isexec(&(filelist[filepos])))
 		? strlen(filelist[filepos].name) + 1 : 0;
+	ttyiomode(1);
 	com = inputshellloop(len, filelist[filepos].name);
+	ttyiomode(0);
 #else
 	if (isdir(&(filelist[filepos])) || !isexec(&(filelist[filepos]))) {
 		len = 0;
@@ -1279,7 +1330,9 @@ char *arg;
 		tmp[1] = _SC_;
 		strcpy(&(tmp[2]), filelist[filepos].name);
 	}
+	ttyiomode(1);
 	com = inputshellloop(len, tmp);
+	ttyiomode(0);
 	if (tmp != filelist[filepos].name) free(tmp);
 #endif
 	if (!com) return(1);
@@ -1294,7 +1347,7 @@ char *arg;
 		if (!(dir = tmpunpack(1))) ret = -1;
 		else {
 			ret = execusercomm(com,
-				filelist[filepos].name, 0, 1, 1);
+				filelist[filepos].name, 0, -1, 1);
 			removetmp(dir, archivedir, filelist[filepos].name);
 		}
 	}
@@ -1303,13 +1356,13 @@ char *arg;
 #ifndef	_NODOSDRIVE
 	if ((drive = tmpdosdupl("", &dir, 1)) < 0) ret = -1;
 	else if (drive) {
-		ret = execusercomm(com, filelist[filepos].name, 0, 1, 1);
+		ret = execusercomm(com, filelist[filepos].name, 0, -1, 1);
 		removetmp(dir, NULL, filelist[filepos].name);
 	}
 	else
 #endif
 	{
-		ret = execusercomm(com, filelist[filepos].name, 0, 1, 0);
+		ret = execusercomm(com, filelist[filepos].name, 0, -1, 0);
 	}
 	ret = (ret < 0) ? 1 : (internal_status >= -1) ? internal_status: 4;
 	free(com);
@@ -1320,50 +1373,24 @@ char *arg;
 static int launch_file(arg)
 char *arg;
 {
-	if (launcher() < 0) return(view_file(arg));
+	if (!launcher()) return(view_file(arg));
 	return(4);
 }
 
 static int pack_file(arg)
 char *arg;
 {
-#if	!defined (_NOARCHIVE) || !defined (_NODOSDRIVE)
-	char *dir;
-#endif
-#ifndef	_NODOSDRIVE
-	int drive;
-#endif
 	char *file;
 	int ret;
 
-#if	!defined (_NOARCHIVE) || !defined (_NODOSDRIVE)
-	dir = NULL;
-#endif
-#ifndef	_NODOSDRIVE
-	drive = 0;
-#endif
 	if (arg && *arg) file = strdup2(arg);
-	else if (!(file = inputstr(PACK_K, 1, -1, NULL, 1))
-	|| !*(file = evalpath(file, 1))
-#ifndef	_NOARCHIVE
-	|| (archivefile && !(dir = tmpunpack(0)))
-#endif
-#ifndef	_NODOSDRIVE
-	|| (drive = tmpdosdupl("", &dir, 0)) < 0
-#endif
-	) return(1);
+	else if (!(file = inputstr(PACK_K, 1, -1, NULL, 1))) return(1);
+	else if (!*(file = evalpath(file, 1))) {
+		free(file);
+		return(1);
+	}
 
 	ret = pack(file);
-#ifndef	_NODOSDRIVE
-	if (drive) removetmp(dir, NULL, "");
-	else
-#endif
-#ifndef	_NOARCHIVE
-	if (archivefile) removetmp(dir, archivedir, "");
-	else
-#endif
-	;
-
 	free(file);
 	if (ret < 0) {
 		putterm(t_bell);
@@ -1378,17 +1405,14 @@ char *arg;
 {
 	int ret;
 
-	if (isdir(&(filelist[filepos]))) return(warning_bell(arg));
 	if (archivefile) ret = unpack(archivefile, NULL, arg, 0, 0);
+	else if (isdir(&(filelist[filepos]))) return(warning_bell(arg));
 	else {
-#if	MSDOS || defined (_NODOSDRIVE)
-		ret = unpack(filelist[filepos].name, NULL, arg, 0, 1);
-#else
+#if	!MSDOS && !defined (_NODOSDRIVE)
 		char path[MAXPATHLEN];
-
-		ret = unpack(nodospath(path, filelist[filepos].name),
-			NULL, arg, 0, 1);
 #endif
+
+		ret = unpack(fnodospath(path, filepos), NULL, arg, 0, 1);
 	}
 	if (ret < 0) return(warning_bell(arg));
 	if (!ret) return(1);
@@ -1401,17 +1425,14 @@ char *arg;
 {
 	int ret;
 
-	if (isdir(&(filelist[filepos]))) return(warning_bell(arg));
 	if (archivefile) ret = unpack(archivefile, NULL, NULL, 1, 0);
+	else if (isdir(&(filelist[filepos]))) return(warning_bell(arg));
 	else {
-#if	MSDOS || defined (_NODOSDRIVE)
-		ret = unpack(filelist[filepos].name, NULL, NULL, 1, 1);
-#else
+#if	!MSDOS && !defined (_NODOSDRIVE)
 		char path[MAXPATHLEN];
-
-		ret = unpack(nodospath(path, filelist[filepos].name),
-			NULL, NULL, 1, 1);
 #endif
+
+		ret = unpack(fnodospath(path, filepos), NULL, NULL, 1, 1);
 	}
 	if (ret <= 0) {
 		if (ret < 0) warning_bell(arg);
@@ -1428,12 +1449,13 @@ char *arg;
 	char *path;
 	int ret;
 
-	if (arg && *arg) path = evalpath(strdup2(arg), 1);
+	if (arg && *arg) path = strdup2(arg);
 	else if (!(path = inputstr(FSDIR_K, 0, -1, NULL, 1))) return(1);
 	else if (!*(path = evalpath(path, 1))) {
 		free(path);
 		path = strdup2(".");
 	}
+
 	ret = infofs(path);
 	free(path);
 	if (!ret) return(1);
@@ -1457,7 +1479,7 @@ char *arg;
 			if (ismark(&(filelist[i]))) break;
 		if (i >= maxfile) i = filepos;
 		flag = 1;
-		locate(0, LINFO);
+		locate(0, L_INFO);
 		putterm(l_clear);
 		kanjiputs(ATTRM_K);
 		if (selectstr(&flag, 2, 35, str, val) == K_ESC) return(1);
@@ -1474,7 +1496,10 @@ char *arg;
 	}
 
 	while ((n = inputattr(&(filelist[i]), flag)) < 0) warning(0, ILTMS_K);
-	if (!n) return(2);
+	if (!n) {
+		if (FILEPERROW < WFILEMINATTR) return(4);
+		return(2);
+	}
 
 	applyfile(setattr, NULL);
 	return(4);
@@ -1505,46 +1530,45 @@ char *arg;
 static int backup_tape(arg)
 char *arg;
 {
-#if	!defined (_NOARCHIVE) || !defined (_NODOSDRIVE)
-	char *dir;
-#endif
+	char *dir, *dev;
 #ifndef	_NODOSDRIVE
 	int drive;
 #endif
-	char *dev;
 	int ret;
 
-#if	!defined (_NOARCHIVE) || !defined (_NODOSDRIVE)
 	dir = NULL;
-#endif
 #ifndef	_NODOSDRIVE
 	drive = 0;
 #endif
 	if (arg && *arg) dev = strdup2(arg);
 #if	MSDOS
-	else if (!(dev = inputstr(BKUP_K, 1, -1, NULL, 1))
+	else if (!(dev = inputstr(BKUP_K, 1, -1, NULL, 1))) return(1);
 #else
-	else if (!(dev = inputstr(BKUP_K, 1, 5, "/dev/", 1))
-#endif
-	|| !*(dev = evalpath(dev, 1))
-#ifndef	_NOARCHIVE
-	|| (archivefile && !(dir = tmpunpack(0)))
+	else if (!(dev = inputstr(BKUP_K, 1, 5, "/dev/", 1))) return(1);
 #endif
+	else if (!*(dev = evalpath(dev, 1))) {
+		free(dev);
+		return(1);
+	}
+	else if (archivefile) {
+		if (!(dir = tmpunpack(0))) {
+			free(dev);
+			return(1);
+		}
+	}
 #ifndef	_NODOSDRIVE
-	|| (drive = tmpdosdupl("", &dir, 0)) < 0
+	else if ((drive = tmpdosdupl("", &dir, 0)) < 0) {
+		free(dev);
+		return(1);
+	}
 #endif
-	) return(1);
 
 	ret = backup(dev);
 #ifndef	_NODOSDRIVE
 	if (drive) removetmp(dir, NULL, "");
 	else
 #endif
-#ifndef	_NOARCHIVE
 	if (archivefile) removetmp(dir, archivedir, "");
-	else
-#endif
-	;
 
 	free(dev);
 	if (ret <= 0) return(1);
@@ -1557,11 +1581,11 @@ static int search_forw(arg)
 char *arg;
 {
 	isearch = 1;
-	locate(0, LHELP);
+	locate(0, L_HELP);
 	putterm(l_clear);
 	putterm(t_standout);
 	win_x = kanjiputs(SEAF_K);
-	win_y = LHELP;
+	win_y = L_HELP;
 	putterm(end_standout);
 	return(0);
 }
@@ -1571,11 +1595,11 @@ static int search_back(arg)
 char *arg;
 {
 	isearch = -1;
-	locate(0, LHELP);
+	locate(0, L_HELP);
 	putterm(l_clear);
 	putterm(t_standout);
 	win_x = kanjiputs(SEAB_K);
-	win_y = LHELP;
+	win_y = L_HELP;
 	putterm(end_standout);
 	return(0);
 }
@@ -1624,6 +1648,12 @@ char *arg;
 	oldwin = win;
 	if (windows < MAXWINDOWS) {
 		win = windows++;
+		if (FILEPERROW < WFILEMIN) {
+			windows--;
+			win = oldwin;
+			warning(0, NOROW_K);
+			return(1);
+		}
 		duplwin(oldwin);
 		movewin(oldwin);
 	}
@@ -1653,7 +1683,7 @@ char *arg;
 	if (chdir2(fullpath) < 0) lostcwd(fullpath);
 	return(4);
 }
-#endif
+#endif	/* !_NOSPLITWIN */
 
 /*ARGSUSED*/
 static int warning_bell(arg)
