@@ -18,14 +18,13 @@
 #include <process.h>
 extern int getcurdrv __P_((VOID_A));
 # ifndef	_NOUSELFN
-extern char *preparefile __P_((char *, char *, int));
+extern char *preparefile __P_((char *, char *));
 # endif
 #else
 #include <pwd.h>
 #include <grp.h>
-#include <sys/file.h>
-#include <sys/param.h>
 #endif
+
 #ifndef	_NODOSDRIVE
 extern int flushdrv __P_((int, VOID_T (*)__P_((VOID_A))));
 extern int shutdrv __P_((int));
@@ -41,13 +40,16 @@ extern off_t totalsize;
 extern off_t blocksize;
 extern char fullpath[];
 extern char *tmpfilename;
+extern int physical_path;
 
-#define	DOSDIRENT		32
-#define	LFNENTSIZ		13
 #define	MAXTMPNAMLEN		8
 #ifndef	O_BINARY
 #define	O_BINARY		0
 #endif
+#ifdef	_NODOSDRIVE
+#define	DOSDIRENT	32
+#define	LFNENTSIZ	13
+#endif
 
 #define	realdirsiz(name)	((dos) ? ((k_strlen(name) / LFNENTSIZ + 1) \
 					* DOSDIRENT + DOSDIRENT) \
@@ -60,12 +62,15 @@ extern char *tmpfilename;
 
 static int NEAR cpfile __P_((char *, char *, struct stat *, struct stat *));
 static VOID changemes __P_((VOID_A));
+static int NEAR genrand __P_((int));
+static int dounlink __P_((char *));
 #ifndef	_NOWRITEFS
+static int NEAR isexist __P_((char *));
 static int NEAR k_strlen __P_((char *));
 static int NEAR saferename __P_((char *, char *));
-static char *NEAR maketmpfile __P_((int, int, char *, char *));
+static char *NEAR maketmpfile __P_((int, char *, char *));
 #if	!MSDOS
-static char *NEAR getentnum __P_((char *, int, int));
+static char *NEAR getentnum __P_((char *, long, int));
 #endif
 static VOID NEAR restorefile __P_((char *, char *, int));
 #endif	/* !_NOWRITEFS */
@@ -74,6 +79,7 @@ char *deftmpdir = NULL;
 #if	!MSDOS && !defined (_NOEXTRACOPY)
 int inheritcopy = 0;
 #endif
+int tmpumask = TMPUMASK;
 
 #ifndef	_NODOSDRIVE
 static int dosdrv = -1;
@@ -87,7 +93,7 @@ char *path, *file;
 	if (!_dospath(file)) return(file);
 	path[0] = '.';
 	path[1] = _SC_;
-	strcpy(path + 2, file);
+	strcpy(&(path[2]), file);
 	return(path);
 }
 #endif
@@ -174,20 +180,20 @@ gid_t gid;
 int getstatus(list)
 namelist *list;
 {
+#if	!MSDOS
+# ifndef	_NODOSDRIVE
+	char path[MAXPATHLEN];
+# endif
+	char *cp;
+#endif
 	struct stat st, lst;
 
 #if	MSDOS
 	if (stat2(list -> name, &st) < 0) return(-1);
 	memcpy((char *)&lst, (char *)&st, sizeof(struct stat));
 #else	/* !MSDOS */
-	char *cp;
-
-# ifdef	_NODOSDRIVE
-	cp = list -> name;
-# else
-	char path[MAXPATHLEN];
-
 	cp = nodospath(path, list -> name);
+# ifndef	_NODOSDRIVE
 	if (dospath(cp, NULL)) {
 		if (stat2(cp, &st) < 0) return(-1);
 		memcpy((char *)&lst, (char *)&st, sizeof(struct stat));
@@ -235,14 +241,6 @@ namelist *list;
 	return(0);
 }
 
-int isdotdir(name)
-char *name;
-{
-	if (name[0] == '.'
-	&& (!name[1] || (name[1] == '.' && !name[2]))) return(1);
-	return(0);
-}
-
 int cmplist(vp1, vp2)
 CONST VOID_P vp1;
 CONST VOID_P vp2;
@@ -274,7 +272,7 @@ CONST VOID_P vp2;
 		case 5:
 			tmp = (int)strlen(list1 -> name)
 				- (int)strlen(list2 -> name);
-			if (tmp != 0) break;
+			if (tmp != 0L) break;
 /*FALLTHRU*/
 		case 1:
 			tmp = strpathcmp2(list1 -> name, list2 -> name);
@@ -297,7 +295,7 @@ CONST VOID_P vp2;
 					break;
 				}
 				else if (*cp1 != '.') {
-					tmp = -1;
+					tmp = -1L;
 					break;
 				}
 				if ((tmp = strpathcmp2(cp1 + 1, cp2 + 1)))
@@ -315,15 +313,15 @@ CONST VOID_P vp2;
 			tmp = list1 -> st_mtim - list2 -> st_mtim;
 			break;
 		default:
-			tmp = 0;
+			tmp = 0L;
 			break;
 	}
 
 	if (sorton > 7) tmp = -tmp;
 	if (!tmp) tmp = list1 -> ent - list2 -> ent;
 
-	if (tmp > 0) return(1);
-	if (tmp < 0) return(-1);
+	if (tmp > 0L) return(1);
+	if (tmp < 0L) return(-1);
 	return(0);
 }
 
@@ -421,6 +419,7 @@ char *buf;
 		if (_chdir2(cwd) < 0) error(cwd);
 	}
 
+	if (buf && !physical_path) strcpy(cwd, fullpath);
 	len = strlen(homedir);
 #if	MSDOS
 	if (len <= 3 || strnpathcmp(cwd, homedir, len)) {
@@ -446,32 +445,36 @@ char *dir;
 {
 	struct stat st;
 	char tmp[MAXPATHLEN];
-#if	MSDOS
 	char *cp;
-#endif
 
-	if (stat2(dir, &st) < 0) {
-		if (errno != ENOENT) return(-1);
-		if (Xmkdir(dir, 0777) < 0) {
-#if	MSDOS
-			if (errno != EEXIST) return(-1);
-			cp = (_dospath(dir)) ? dir + 2 : dir;
-			if (cp[0] != _SC_ || cp[1]) {
-				if ((cp = strrdelim(dir, 1))) cp++;
-				else cp = dir;
-				if (cp[0] != '.' || cp[1]) return(-1);
-			}
-			st.st_mode = S_IFDIR;
-#else
-			return(-1);
+	cp = dir;
+#if	MSDOS || !defined (_NODOSDRIVE)
+	if (_dospath(dir)) cp += 2;
 #endif
+	if (!isdotdir(cp)) {
+		if (stat2(dir, &st) < 0) {
+			if (errno != ENOENT) return(-1);
+			if (Xmkdir(dir, 0777) < 0) {
+#if	MSDOS
+				if (errno != EEXIST) return(-1);
+				if (cp[0] != _SC_ || cp[1]) {
+					if ((cp = strrdelim(dir, 1))) cp++;
+					else cp = dir;
+					if (cp[0] != '.' || cp[1]) return(-1);
+				}
+				st.st_mode = S_IFDIR;
+#else
+				return(-1);
+#endif
+			}
+			else if (stat2(dir, &st) < 0) return(-1);
+		}
+		if ((st.st_mode & S_IFMT) != S_IFDIR) {
+			errno = ENOTDIR;
+			return(-1);
 		}
-		else if (stat2(dir, &st) < 0) return(-1);
-	}
-	if ((st.st_mode & S_IFMT) != S_IFDIR) {
-		errno = ENOTDIR;
-		return(-1);
 	}
+
 	entryhist(1, realpath2(dir, tmp, 0), 1);
 	return(0);
 }
@@ -487,6 +490,9 @@ struct stat *stp;
 	if ((st.st_mode & S_IFMT) == S_IFLNK) return(1);
 
 	ret = 0;
+#ifdef	FAKEUNINIT
+	duperrno = errno;	/* fake for -Wuninitialized */
+#endif
 #if	MSDOS
 	if (!(st.st_mode & S_IWRITE)) Xchmod(path, (st.st_mode | S_IWRITE));
 #endif
@@ -513,11 +519,11 @@ struct stat *stp;
 		}
 #endif
 	}
-#ifdef HAVEFLAGS
+#ifdef	HAVEFLAGS
 	if (stp -> st_flags != 0xffffffff) {
 # ifndef	_NODOSDRIVE
 		if (dospath(path, NULL)) {
-			errno = EACCES;
+			duperrno = EACCES;
 			ret = -1;
 		}
 		else
@@ -541,6 +547,10 @@ struct stat *stp;
 	else if (!(st.st_mode & S_IWRITE)) Xchmod(path, st.st_mode);
 #else
 	if (stp -> st_gid != (gid_t)-1) {
+# ifndef	_NODOSDRIVE
+		if (dospath(path, NULL));
+		else
+# endif
 		if (chown(path, stp -> st_uid, stp -> st_gid) < 0)
 			chown(path, -1, stp -> st_gid);
 	}
@@ -558,16 +568,14 @@ char *path;
 	duperrno = errno;
 	if (!path) path = buf;
 
-	if (path != fullpath && !rawchdir(fullpath) && _Xgetwd(path))
+	if (path != fullpath && !rawchdir(fullpath) && _Xgetwd(path, 1))
 		cp = NOCWD_K;
-	else {
-		if ((cp = gethomedir()) && !rawchdir(cp) && _Xgetwd(path))
-			cp = GOHOM_K;
-		else if (!rawchdir(_SS_) && _Xgetwd(path)) cp = GOROT_K;
-		else error(_SS_);
-		strncpy2(fullpath, path, MAXPATHLEN - 1);
-	}
+	else if ((cp = gethomedir()) && !rawchdir(cp) && _Xgetwd(path, 1))
+		cp = GOHOM_K;
+	else if (!rawchdir(_SS_) && _Xgetwd(path, 1)) cp = GOROT_K;
+	else error(_SS_);
 
+	if (path != fullpath) strncpy2(fullpath, path, MAXPATHLEN - 1);
 	warning(0, cp);
 	errno = duperrno;
 }
@@ -581,12 +589,12 @@ int size;
 
 	n = Xwrite(fd, buf, size);
 #if	MSDOS
-	if (n < size) n = 0;
-#else
-	if (n < 0) {
-		if (errno == EINTR) n = size;
-		else if (errno == ENOSPC) n = 0;
+	if (n >= 0 && n < size) {
+		n = -1;
+		errno = ENOSPC;
 	}
+#else
+	if (n < 0 && errno == EINTR) n = size;
 #endif
 	return(n);
 }
@@ -600,7 +608,7 @@ struct stat *stp1, *stp2;
 	struct stat st;
 #endif
 	char buf[BUFSIZ];
-	int i, n, fd1, fd2, tmperrno;
+	int i, fd1, fd2, tmperrno;
 
 #if	MSDOS
 	if (!stp2 && Xlstat(dest, &st) >= 0) stp2 = &st;
@@ -612,14 +620,17 @@ struct stat *stp1, *stp2;
 		buf[i] = '\0';
 		return(Xsymlink(buf, dest));
 	}
-	if ((fd1 = Xopen(src, O_RDONLY | O_BINARY, stp1 -> st_mode)) < 0)
+	if ((fd1 = Xopen(src, O_BINARY | O_RDONLY, stp1 -> st_mode)) < 0)
 		return(-1);
-	if ((fd2 = Xopen(dest, O_WRONLY | O_BINARY | O_CREAT | O_TRUNC,
 #if	MSDOS
-	stp1 -> st_mode | S_IWRITE)) < 0) {
+	fd2 = Xopen(dest,
+		O_BINARY | O_WRONLY | O_CREAT | O_TRUNC,
+		stp1 -> st_mode | S_IWRITE);
 #else
-	stp1 -> st_mode)) < 0) {
+	fd2 = Xopen(dest,
+		O_BINARY | O_WRONLY | O_CREAT | O_TRUNC, stp1 -> st_mode);
 #endif
+	if (fd2 < 0) {
 		Xclose(fd1);
 		return(-1);
 	}
@@ -628,16 +639,12 @@ struct stat *stp1, *stp2;
 		while ((i = Xread(fd1, buf, BUFSIZ)) < 0 && errno == EINTR);
 		tmperrno = errno;
 		if (i < BUFSIZ) break;
-		if ((n = safewrite(fd2, buf, BUFSIZ)) <= 0) {
-			tmperrno = (n < 0) ? errno : ENOSPC;
-			i = -1;
+		if ((i = safewrite(fd2, buf, BUFSIZ)) < 0) {
+			tmperrno = errno;
 			break;
 		}
 	}
-	if (i > 0 && (n = safewrite(fd2, buf, i)) <= 0) {
-		tmperrno = (n < 0) ? errno : ENOSPC;
-		i = -1;
-	}
+	if (i > 0 && (i = safewrite(fd2, buf, i)) < 0) tmperrno = errno;
 
 	Xclose(fd2);
 	Xclose(fd1);
@@ -647,6 +654,9 @@ struct stat *stp1, *stp2;
 		errno = tmperrno;
 		return(-1);
 	}
+#ifdef	HAVEFLAGS
+	stp1 -> st_flags = 0xffffffff;
+#endif
 #if	MSDOS
 	if (touchfile(dest, stp1) < 0) return(-1);
 #else
@@ -697,19 +707,76 @@ struct stat *stp1, *stp2;
 	if (Xrename(src, dest) >= 0) return(0);
 	if (errno != EXDEV || (stp1 -> st_mode & S_IFMT) == S_IFDIR)
 		return(-1);
-	if (safecpfile(src, dest, stp1, stp2) < 0 || Xunlink(src) < 0
-	|| touchfile(dest, stp1) < 0) return(-1);
-	return(0);
+	if (safecpfile(src, dest, stp1, stp2) < 0 || Xunlink(src) < 0)
+		return(-1);
+#ifdef	HAVEFLAGS
+	stp1 -> st_flags = 0xffffffff;
+#endif
+	return (touchfile(dest, stp1));
+}
+
+static int NEAR genrand(max)
+int max;
+{
+	static long last = -1L;
+	time_t now;
+
+	if (last < 0L) {
+		now = time2();
+		last = ((now & 0xff) << 16) + (now & ~0xff) + getpid();
+	}
+
+	do {
+		last = last * (u_long)1103515245 + 12345;
+	} while (last < 0L);
+
+	return((last / 65537L) % max);
+}
+
+char *genrandname(buf, len)
+char *buf;
+int len;
+{
+	static char seq[] = {
+#if	MSDOS
+		'0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
+		'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J',
+		'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T',
+		'U', 'V', 'W', 'X', 'Y', 'Z', '_'
+#else
+		'0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
+		'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j',
+		'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't',
+		'u', 'v', 'w', 'x', 'y', 'z', '_'
+#endif
+	};
+	int i, j, c;
+
+	if (!buf) {
+		for (i = 0; i < sizeof(seq) / sizeof(char); i++) {
+			j = genrand(sizeof(seq) / sizeof(char));
+			c = seq[i];
+			seq[i] = seq[j];
+			seq[j] = c;
+		}
+	}
+	else {
+		for (i = 0; i < len; i++) {
+			j = genrand(sizeof(seq) / sizeof(char));
+			buf[i] = seq[j];
+		}
+		buf[i] = '\0';
+	}
+	return(buf);
 }
 
 int mktmpdir(dir)
 char *dir;
 {
 	char *cp, path[MAXPATHLEN];
-	int no;
+	int n, len, mask;
 
-	if (!deftmpdir || !*deftmpdir || !tmpfilename || !*tmpfilename
-	|| !dir || !*dir) {
+	if (!deftmpdir || !*deftmpdir || !dir) {
 		errno = ENOENT;
 		return(-1);
 	}
@@ -717,7 +784,7 @@ char *dir;
 	free(deftmpdir);
 #if	MSDOS && !defined (_NODOSDRIVE)
 	if (checkdrive(toupper2(path[0]) - 'A'))
-		if (!_Xgetwd(path)) {
+		if (!_Xgetwd(path, 1)) {
 			lostcwd(path);
 			deftmpdir = NULL;
 			return(-1);
@@ -725,23 +792,62 @@ char *dir;
 #endif
 	deftmpdir = strdup2(path);
 #if	MSDOS
-	no = getcurdrv();
-	*path = (no >= 'A' && no <= 'Z') ? toupper2(*path) : tolower2(*path);
+	n = getcurdrv();
+	*path = (n >= 'A' && n <= 'Z') ? toupper2(*path) : tolower2(*path);
 #endif
-	strcpy(strcatdelim(path), tmpfilename);
-	if (_Xmkdir(path, 0755) < 0 && errno != EEXIST) return(-1);
-	strcpy((cp = strcatdelim(path)), dir);
-	if (_Xmkdir(path, 0755) < 0 && errno != EEXIST) {
+
+	mask = 0777 & ~tmpumask;
+	cp = strcatdelim(path);
+	if (tmpfilename && Xaccess(tmpfilename, R_OK | W_OK | X_OK) < 0) {
+		free(tmpfilename);
+		tmpfilename = NULL;
+	}
+	if (tmpfilename) strcpy(cp, tmpfilename);
+	else {
+		n = sizeof(TMPPREFIX) - 1;
+		strncpy(cp, TMPPREFIX, n);
+		len = sizeof(path) - 1 - (cp - path);
+		if (len > MAXTMPNAMLEN) len = MAXTMPNAMLEN;
+		len -= n;
+		genrandname(NULL, 0);
+
+		for (;;) {
+			genrandname(cp + n, len);
+			if (_Xmkdir(path, mask, 0, 1) >= 0) break;
+			if (errno != EEXIST) return(-1);
+		}
+		tmpfilename = strdup2(cp);
+	}
+
+	if (!(n = strlen(dir))) {
+		strcpy(dir, path);
+		return(0);
+	}
+
+	strncpy((cp = strcatdelim(path)), dir, n);
+	len = sizeof(path) - 1 - (cp - path);
+	if (len > MAXTMPNAMLEN) len = MAXTMPNAMLEN;
+	len -= n;
+	genrandname(NULL, 0);
+
+	for (;;) {
+		genrandname(cp + n, len);
+		if (_Xmkdir(path, mask, 0, 1) >= 0) {
+			strcpy(dir, path);
+			return(0);
+		}
+		if (errno != EEXIST) break;
+	}
+
+	if (cp > path) {
 		*(--cp) = '\0';
-		no = errno;
-		if (_Xrmdir(path) < 0
+		n = errno;
+		if (Xrmdir(path) < 0
 		&& errno != ENOTEMPTY && errno != EEXIST && errno != EACCES)
 			error(path);
-		errno = no;
-		return(-1);
+		errno = n;
 	}
-	strcpy(dir, path);
-	return(0);
+	return(-1);
 }
 
 int rmtmpdir(dir)
@@ -749,86 +855,77 @@ char *dir;
 {
 	char path[MAXPATHLEN];
 
-	if (dir && *dir && _Xrmdir(dir) < 0) return(-1);
+	if (dir && *dir && _Xrmdir(dir, 0, 1) < 0) return(-1);
 	if (!deftmpdir || !*deftmpdir || !tmpfilename || !*tmpfilename) {
 		errno = ENOENT;
 		return(-1);
 	}
 	strcatdelim2(path, deftmpdir, tmpfilename);
-	if (_Xrmdir(path) < 0
-	&& errno != ENOTEMPTY && errno != EEXIST && errno != EACCES)
+	if (_Xrmdir(path, 0, 1) >= 0) {
+		free(tmpfilename);
+		tmpfilename = NULL;
+	}
+	else if (errno != ENOTEMPTY && errno != EEXIST && errno != EACCES)
 		return(-1);
 	return(0);
 }
 
-int mktmpfile(buf, dir)
-char *buf, *dir;
+int mktmpfile(file)
+char *file;
 {
 	char *cp, path[MAXPATHLEN];
-	int i, fd;
+	int fd, len, duperrno;
 
-	strcpy(path, dir);
+	*path = '\0';
 	if (mktmpdir(path) < 0) return(-1);
 	cp = strcatdelim(path);
-	memset(cp, '\0', sizeof(path) - (cp - path));
-	cp[0] = '_';
+	len = sizeof(path) - 1 - (cp - path);
+	if (len > MAXTMPNAMLEN) len = MAXTMPNAMLEN;
+	genrandname(NULL, 0);
 
 	for (;;) {
-		fd = Xopen(path, O_BINARY | O_WRONLY | O_CREAT | O_EXCL, 0644);
+		genrandname(cp, len);
+		fd = Xopen(path, O_BINARY | O_WRONLY | O_CREAT | O_EXCL,
+			0666 & ~tmpumask);
 		if (fd >= 0) {
-			strcpy(buf, path);
+			strcpy(file, path);
 			return(fd);
 		}
-		if (errno != EEXIST) return(-1);
-
-		for (i = 0; i < MAXTMPNAMLEN; i++) {
-			if (!cp[i]) cp[i] = '_';
-			else if (cp[i] == '9') {
-				cp[i] = '_';
-				continue;
-			}
-			else if (cp[i] == '_') cp[i] = 'a';
-#if	MSDOS
-			else if (cp[i] == 'z') cp[i] = '0';
-#else
-			else if (cp[i] == 'z') cp[i] = 'A';
-			else if (cp[i] == 'Z') cp[i] = '0';
-#endif
-			else cp[i]++;
-			break;
-		}
-		if (i >= MAXTMPNAMLEN) break;
+		if (errno != EEXIST) break;
 	}
-	if (cp > path) *(--cp) = '\0';
-	rmtmpdir(path);
+	duperrno = errno;
+	rmtmpdir(NULL);
+	errno = duperrno;
 	return(-1);
 }
 
-int rmtmpfile(path)
+int rmtmpfile(file)
+char *file;
+{
+	if ((Xunlink(file) < 0 && errno != ENOENT) || rmtmpdir(NULL) < 0)
+		return(-1);
+	return(0);
+}
+
+static int dounlink(path)
 char *path;
 {
-	char *cp;
-	int ret;
+	return(Xunlink(path));
+}
 
-	ret = 0;
-	if (Xunlink(path) != 0 && errno != ENOENT) ret = -1;
-	else if (!(cp = strrdelim(path, 0)));
-#if	MSDOS
-	else if (cp == path + 2 && isalpha(path[0]) && path[1] == ':');
-#endif
-	else if (cp != path) {
-		*cp = '\0';
-		if (rmtmpdir(path) < 0
-		&& errno != ENOTEMPTY && errno != EEXIST && errno != EACCES)
-			ret = -1;
-	}
-	free(path);
-	return(ret);
+static int dormdir(path)
+char *path;
+{
+	if (isdotdir(path)) return(0);
+	return(Xrmdir(path));
 }
 
 VOID removetmp(dir, subdir, file)
 char *dir, *subdir, *file;
 {
+#if	!MSDOS && !defined (_NODOSDRIVE)
+	char path[MAXPATHLEN];
+#endif
 	char *cp, *tmp, *dupdir;
 
 	if (!dir || !*dir) subdir = file = NULL;
@@ -846,20 +943,13 @@ char *dir, *subdir, *file;
 	}
 	if (file) {
 		if (*file) {
-#if	!MSDOS && !defined (_NODOSDRIVE)
-			char path[MAXPATHLEN];
-
-			file = nodospath(path, file);
-#endif
-			if (Xunlink(file) < 0) {
+			if (Xunlink(nodospath(path, file)) < 0) {
 				warning(-1, file);
 				subdir = NULL;
 			}
 		}
-		else if (applydir(".", Xunlink, NULL, NULL, 0, NULL) < 0) {
-			warning(-1, dir);
+		else if (applydir(NULL, dounlink, NULL, dormdir, 3, NULL) < 0)
 			subdir = NULL;
-		}
 	}
 	if (subdir && *subdir) {
 		if (_chdir2(dir) < 0) error(dir);
@@ -869,7 +959,8 @@ char *dir, *subdir, *file;
 			*cp = '\0';
 			if ((tmp = strrdelim(dupdir, 1))) tmp++;
 			else tmp = dupdir;
-			if ((*tmp && strcmp(tmp, ".") && _Xrmdir(dupdir) < 0)
+			if ((*tmp && strcmp(tmp, ".")
+			&& _Xrmdir(dupdir, 0, 1) < 0)
 			|| tmp == dupdir) break;
 			cp = tmp - 1;
 		}
@@ -896,7 +987,7 @@ int forcecleandir(dir, file)
 char *dir, *file;
 {
 #if	!MSDOS
-	long pid;
+	p_id_t pid;
 #endif
 	extern char **environ;
 	char buf[MAXPATHLEN];
@@ -909,7 +1000,7 @@ char *dir, *file;
 #if	MSDOS
 	spawnlpe(P_WAIT, "DELTREE.EXE", "DELTREE", "/Y", buf, NULL, environ);
 #else
-	if ((pid = fork()) < 0) return(-1);
+	if ((pid = fork()) < 0L) return(-1);
 	else if (!pid) {
 		execle("/bin/rm", "rm", "-rf", buf, NULL, environ);
 		_exit(1);
@@ -919,27 +1010,36 @@ char *dir, *file;
 }
 
 #ifndef	_NODOSDRIVE
+char *dostmpdir(drive)
+int drive;
+{
+	char path[MAXPATHLEN];
+
+	path[0] = DOSTMPPREFIX;
+	path[1] = drive;
+	path[2] = '\0';
+	if (mktmpdir(path) < 0) return(NULL);
+	return(strdup2(path));
+}
+
 int tmpdosdupl(dir, dirp, single)
 char *dir, **dirp;
 int single;
 {
 	struct stat st;
-	char *cp, path[MAXPATHLEN];
+	char *cp, *tmpdir, path[MAXPATHLEN];
 # if	!MSDOS
 	char tmp[MAXPATHLEN];
 # endif
 	int i, drive;
 
 	if (!(drive = dospath2(dir))) return(0);
-	cp = path;
-	*(cp++) = 'D';
-	*(cp++) = drive;
-	*cp = '\0';
-	if (mktmpdir(path) < 0) {
-		warning(-1, path);
+	if (!(tmpdir = dostmpdir(drive))) {
+		warning(-1, dir);
 		return(-1);
 	}
 
+	strcpy(path, tmpdir);
 	cp = strcatdelim(path);
 	waitmes();
 
@@ -947,15 +1047,9 @@ int single;
 		strcpy(cp, filelist[filepos].name);
 		st.st_mode = filelist[filepos].st_mode;
 		st.st_atime = st.st_mtime = filelist[filepos].st_mtim;
-# if	MSDOS
-		if (cpfile(filelist[filepos].name, path, &st, NULL) < 0)
-# else
-		if (cpfile(nodospath(tmp, filelist[filepos].name),
-		path, &st, NULL) < 0)
-# endif
-		{
-			*(--cp) = '\0';
-			removetmp(path, NULL, NULL);
+		if (cpfile(fnodospath(tmp, filepos), path, &st, NULL) < 0) {
+			warning(-1, path);
+			removetmp(tmpdir, NULL, NULL);
 			return(-1);
 		}
 	}
@@ -964,24 +1058,17 @@ int single;
 		strcpy(cp, filelist[i].name);
 		st.st_mode = filelist[i].st_mode;
 		st.st_atime = st.st_mtime = filelist[i].st_mtim;
-# if	MSDOS
-		if (cpfile(filelist[i].name, path, &st, NULL) < 0)
-# else
-		if (cpfile(nodospath(tmp, filelist[i].name),
-		path, &st, NULL) < 0)
-# endif
-		{
-			*(--cp) = '\0';
-			removetmp(path, NULL, "");
+		if (cpfile(fnodospath(tmp, i), path, &st, NULL) < 0) {
+			warning(-1, path);
+			removetmp(tmpdir, NULL, "");
 			return(-1);
 		}
 	}
 
-	*(--cp) = '\0';
 	dosdrv = lastdrv;
 	lastdrv = -1;
-	if (_chdir2(path) < 0) error(path);
-	*dirp = strdup2(path);
+	if (_chdir2(tmpdir) < 0) error(tmpdir);
+	*dirp = tmpdir;
 	return(drive);
 }
 
@@ -997,7 +1084,7 @@ char *file;
 
 	path[0] = drive;
 	path[1] = ':';
-	strcpy(path + 2, file);
+	strcpy(&(path[2]), file);
 # if	!MSDOS
 	file = nodospath(tmp, file);
 # endif
@@ -1009,20 +1096,16 @@ char *file;
 }
 #endif	/* !_NODOSDRIVE */
 
-int isexist(file)
+#ifndef	_NOWRITEFS
+static int NEAR isexist(file)
 char *file;
 {
-#if	MSDOS && !defined (_NOUSELFN)
-	char *cp, buf[MAXPATHLEN];
+	struct stat st;
 
-	if (!(cp = preparefile(file, buf, 0)) && errno == ENOENT) return(0);
-	if (cp != file) return(1);
-#endif
-	if (Xaccess(file, F_OK) < 0 && errno == ENOENT) return(0);
+	if (_Xlstat(file, &st, 1, 0) < 0 && errno == ENOENT) return(0);
 	return(1);
 }
 
-#ifndef	_NOWRITEFS
 static int NEAR k_strlen(s)
 char *s;
 {
@@ -1044,27 +1127,29 @@ char *from, *to;
 	from = nodospath(fpath, from);
 	to = nodospath(tpath, to);
 #endif
-	return(Xrename(from, to));
+	return(_Xrename(from, to, 1));
 }
 
-static char *NEAR maketmpfile(len, dos, tmpdir, old)
-int len, dos;
+static char *NEAR maketmpfile(len, tmpdir, old)
+int len;
 char *tmpdir, *old;
 {
 	char *fname, path[MAXPATHLEN];
-	int i, l, fd;
+	int l, fd;
 
 	if (len < 0) return(NULL);
 	fname = malloc2(len + 1);
-
-	memset(fname, '_', len);
-	fname[len] = '\0';
+	genrandname(NULL, 0);
 
 	if (tmpdir) l = strcatdelim2(path, tmpdir, NULL) - path;
+#ifdef	FAKEUNINIT
+	else l = 0;	/* fake for -Wuninitialized */
+#endif
 
 	for (;;) {
+		genrandname(fname, len);
 		if (tmpdir) {
-			strcpy(path + l, fname);
+			strcpy(&(path[l]), fname);
 			if (!isexist(path)) {
 				if (old) {
 #if	!MSDOS
@@ -1079,8 +1164,8 @@ char *tmpdir, *old;
 					else break;
 				}
 				else {
-					if ((fd = Xopen(fname,
-					O_CREAT | O_EXCL, 0600))) {
+					if ((fd = _Xopen(fname,
+					O_CREAT | O_EXCL, 0600, 1))) {
 						Xclose(fd);
 						return(fname);
 					}
@@ -1089,28 +1174,9 @@ char *tmpdir, *old;
 			}
 		}
 		else {
-			if (_Xmkdir(fname, 0777) >= 0) return(fname);
+			if (_Xmkdir(fname, 0777, 1, 0) >= 0) return(fname);
 			if (errno != EEXIST) break;
 		}
-
-		for (i = len - 1; i >= 0; i--) {
-			if (fname[i] == '9') {
-				fname[i] = '_';
-				continue;
-			}
-			else if (fname[i] == '_') fname[i] = 'a';
-			else if (dos) {
-				if (fname[i] == 'z') fname[i] = '0';
-				else fname[i]++;
-			}
-			else {
-				if (fname[i] == 'z') fname[i] = 'A';
-				else if (fname[i] == 'Z') fname[i] = '0';
-				else fname[i]++;
-			}
-			break;
-		}
-		if (i < 0) break;
 	}
 	free(fname);
 	return(NULL);
@@ -1119,7 +1185,8 @@ char *tmpdir, *old;
 #if	!MSDOS
 static char *NEAR getentnum(dir, bsiz, fs)
 char *dir;
-int bsiz, fs;
+long bsiz;
+int fs;
 {
 	FILE *fp;
 	struct stat st;
@@ -1128,7 +1195,7 @@ int bsiz, fs;
 
 	if (fs != 2) return(NULL);	/* without IRIX File System */
 
-	if (_Xlstat(dir, &st) < 0) error(dir);
+	if (_Xlstat(dir, &st, 1, 0) < 0) error(dir);
 	n = (long)(st.st_size) / bsiz;
 	tmp = malloc2(n + 1);
 
@@ -1151,16 +1218,16 @@ int fnamp;
 	DIR *dirp;
 	struct dirent *dp;
 
-	if (!(dirp = _Xopendir(dir))) error(dir);
-	while ((dp = Xreaddir(dirp))) {
+	if (!(dirp = _Xopendir(dir, 1))) error(dir);
+	while ((dp = _Xreaddir(dirp, 1))) {
 		if (isdotdir(dp -> d_name)) continue;
 		else {
-			strcpy(path + fnamp, dp -> d_name);
+			strcpy(&(path[fnamp]), dp -> d_name);
 			if (saferename(path, dp -> d_name) < 0) error(path);
 		}
 	}
 	Xclosedir(dirp);
-	if (_Xrmdir(dir) < 0) error(dir);
+	if (_Xrmdir(dir, 1, 0) < 0) error(dir);
 	free(dir);
 }
 
@@ -1173,6 +1240,9 @@ int fs;
 	int i, top, len, fnamp, ent;
 	char *cp, *tmpdir, path[MAXPATHLEN];
 #if	!MSDOS
+# ifndef	_NOKANJIFCONV
+	char conv[MAXNAMLEN + 1];
+# endif
 	long persec;
 	char *entnum, **tmpfile;
 	int tmpno, block, totalent, ptr, totalptr;
@@ -1217,8 +1287,14 @@ int fs;
 	if (i >= maxfile) return;
 	top = i;
 	if (dos == 1) len = 8;
-	else len = getnamlen(realdirsiz(filelist[top].name));
-	if (!(tmpdir = maketmpfile(len, dos, NULL, NULL))) {
+	else {
+		cp = filelist[top].name;
+#if	!MSDOS && !defined (_NOKANJIFCONV)
+		cp = kanjiconv2(conv, cp, MAXNAMLEN, DEFCODE, getkcode(cp));
+#endif
+		len = getnamlen(realdirsiz(cp));
+	}
+	if (!(tmpdir = maketmpfile(len, NULL, NULL))) {
 		warning(0, NOWRT_K);
 		return;
 	}
@@ -1228,12 +1304,12 @@ int fs;
 	fnamp = strcatdelim2(path, tmpdir, NULL) - path;
 	waitmes();
 
-	if (!(dirp = _Xopendir("."))) {
+	if (!(dirp = _Xopendir(".", 1))) {
 		lostcwd(path);
 		return;
 	}
 	i = ent = 0;
-	while ((dp = Xreaddir(dirp))) {
+	while ((dp = _Xreaddir(dirp, 1))) {
 		if (isdotdir(dp -> d_name)) continue;
 		else if (!strpathcmp(dp -> d_name, tmpdir)) {
 #if	MSDOS
@@ -1244,7 +1320,7 @@ int fs;
 			ent = i;
 		}
 		else {
-			strcpy(path + fnamp, dp -> d_name);
+			strcpy(&(path[fnamp]), dp -> d_name);
 			if (saferename(dp -> d_name, path) < 0) {
 				Xclosedir(dirp);
 				warning(-1, dp -> d_name);
@@ -1257,8 +1333,7 @@ int fs;
 	Xclosedir(dirp);
 
 	if (ent > 0) {
-		if (!(cp = maketmpfile(len, dos, tmpdir, tmpdir)))
-			error(tmpdir);
+		if (!(cp = maketmpfile(len, tmpdir, tmpdir))) error(tmpdir);
 		free(tmpdir);
 		tmpdir = cp;
 		strcatdelim2(path, tmpdir, NULL);
@@ -1277,8 +1352,12 @@ int fs;
 
 	for (i = 0; i < maxfile; i++) {
 		if (isdotdir(filelist[i].name) || i == top) continue;
+		cp = filelist[i].name;
 #if	!MSDOS
 		ent = persec - totalent;
+# ifndef	_NOKANJIFCONV
+		cp = kanjiconv2(conv, cp, MAXNAMLEN, DEFCODE, getkcode(cp));
+# endif
 		switch (fs) {
 			case 2:	/* IRIX File System */
  				if (totalptr > ptr + 1) ent -= totalptr;
@@ -1287,25 +1366,25 @@ int fs;
 			case 3:	/* SystemV R3 File System */
 			case 4:	/* MS-DOS File System */
 			case 5:	/* Windows95 File System */
-				ent = realdirsiz(filelist[i].name);
+				ent = realdirsiz(cp);
 				break;
 			default:
 				break;
 		}
-		if (ent < realdirsiz(filelist[i].name)) {
+		if (ent < realdirsiz(cp)) {
 			tmpfile = b_realloc(tmpfile, tmpno, char *);
 			tmpfile[tmpno] =
-				maketmpfile(getnamlen(ent), dos, tmpdir, NULL);
+				maketmpfile(getnamlen(ent), tmpdir, NULL);
 			tmpno++;
 			ptr = 0;
 			totalent = headbyte;
 			if (entnum) totalptr = entnum[++block];
 		}
 #endif	/* !MSDOS */
-		strcpy(path + fnamp, filelist[i].name);
-		if (saferename(path, filelist[i].name) < 0) error(path);
+		strcpy(&(path[fnamp]), cp);
+		if (saferename(path, cp) < 0) error(path);
 #if	!MSDOS
-		totalent += realdirsiz(filelist[i].name);
+		totalent += realdirsiz(cp);
 		ptr++;
 #endif
 	}
@@ -1313,16 +1392,20 @@ int fs;
 	if (entnum) free(entnum);
 #endif
 
-	if (!(cp = maketmpfile(len, dos, tmpdir, tmpdir))) error(tmpdir);
+	if (!(cp = maketmpfile(len, tmpdir, tmpdir))) error(tmpdir);
 	free(tmpdir);
 	tmpdir = cp;
-	strcatdelim2(path, tmpdir, filelist[top].name);
-	if (saferename(path, filelist[top].name) < 0) error(path);
+	cp = filelist[top].name;
+#if	!MSDOS && !defined (_NOKANJIFCONV)
+	cp = kanjiconv2(conv, cp, MAXNAMLEN, DEFCODE, getkcode(cp));
+#endif
+	strcatdelim2(path, tmpdir, cp);
+	if (saferename(path, cp) < 0) error(path);
 	restorefile(tmpdir, path, fnamp);
 
 #if	!MSDOS
 	for (i = 0; i < tmpno; i++) if (tmpfile[i]) {
-		if (unlink(tmpfile[i]) < 0) error(tmpfile[i]);
+		if (_Xunlink(tmpfile[i], 1) < 0) error(tmpfile[i]);
 		free(tmpfile[i]);
 	}
 	free(tmpfile);
