@@ -16,6 +16,8 @@ extern int errno;
 #include "unixemu.h"
 #else
 #include <sys/time.h>
+#include <sys/param.h>
+#include <sys/file.h>
 # ifdef	USEDIRECT
 # include <sys/dir.h>
 #  ifdef	DIRSIZ
@@ -33,6 +35,10 @@ extern int errno;
 #include <time.h>
 #endif
 
+#ifndef	_NODOSDRIVE
+#include "dosdisk.h"
+#endif
+
 #if	(GETTODARGS == 1)
 #define	gettimeofday2(tv, tz)	gettimeofday(tv)
 #else
@@ -51,10 +57,22 @@ extern int errno;
 
 #define	getblock(c)	((((c) + blocksize - 1) / blocksize) * blocksize)
 
+#define	isleftshift()	(n_column < WCOLUMNSTD)
+#define	isrightomit()	(n_column < WCOLUMNSTD - 1)
+#define	ispureshift()	(n_column == WCOLUMNSTD - 1)
+#define	isshortwid()	(n_column < WCOLUMNSTD - 1)
+#define	iswellomit()	(n_column < WCOLUMNSTD - 2)
+#define	ishardomit()	(n_column < WCOLUMNOMIT)
+#define	isbestomit()	(n_column < WCOLUMNHARD)
+
 /* main.c */
 extern VOID error __P_((char *));
+extern VOID checkscreen __P_((int, int));
 extern int sigvecset __P_((int));
-extern VOID title __P_((void));
+extern VOID title __P_((VOID_A));
+#ifndef	_NOCUSTOMIZE
+VOID saveorigenviron __P_((VOID_A));
+#endif
 extern int loadruncom __P_((char *, int));
 
 /* dosemu.c or unixemu.c */
@@ -74,39 +92,24 @@ extern int dospath3 __P_((char *));
 #if	!MSDOS && !defined (_NOKANJIFCONV)
 extern int getkcode __P_((char *));
 #endif
-extern DIR *_Xopendir __P_((char *));
-#ifdef	_NOROCKRIDGE
-#define	Xopendir	_Xopendir
-#else
-extern DIR *Xopendir __P_((char *));
-#endif
+extern DIR *_Xopendir __P_((char *, int));
+#define	Xopendir(p)	_Xopendir(p, 0)
 extern int Xclosedir __P_((DIR *));
-extern struct dirent *Xreaddir __P_((DIR *));
+extern struct dirent *_Xreaddir __P_((DIR *, int));
+#define	Xreaddir(d)	_Xreaddir(d, 0)
 extern VOID Xrewinddir __P_((DIR *));
 #if	MSDOS
 extern int rawchdir __P_((char *));
 #else
 #define	rawchdir	chdir
 #endif
-#if	MSDOS && defined (_NODOSDRIVE)
-#define	_Xchdir		rawchdir
-#else
-extern int _Xchdir __P_((char *));
-#endif
-#ifdef	_NOROCKRIDGE
-#define	Xchdir		_Xchdir
-#else
-extern int Xchdir __P_((char *));
-#endif
-extern char *_Xgetwd __P_((char *));
-extern char *Xgetwd __P_((char *));
+extern int _Xchdir __P_((char *, int));
+#define	Xchdir(p)	_Xchdir(p, 0)
+extern char *_Xgetwd __P_((char *, int));
+#define	Xgetwd(p)	_Xgetwd(p, 0)
 extern int Xstat __P_((char *, struct stat *));
-#if	MSDOS || defined (_NOKANJIFCONV)
-#define	_Xlstat(p, s)	(lstat(p, s) ? -1 : 0)
-#else
-extern int _Xlstat __P_((char *, struct stat *));
-#endif
-extern int Xlstat __P_((char *, struct stat *));
+extern int _Xlstat __P_((char *, struct stat *, int, int));
+#define	Xlstat(p, s)	_Xlstat(p, s, 0, 0)
 extern int Xaccess __P_((char *, int));
 extern int Xsymlink __P_((char *, char *));
 extern int Xreadlink __P_((char *, char *, int));
@@ -116,13 +119,12 @@ extern int Xutime __P_((char *, struct utimbuf *));
 #else
 extern int Xutimes __P_((char *, struct timeval []));
 #endif
-extern int Xunlink __P_((char *));
-extern int Xrename __P_((char *, char *));
-#if	MSDOS && defined (_NOROCKRIDGE) && defined (_NOUSELFN)
-#define	Xopen		open
-#else
-extern int Xopen __P_((char *, int, int));
-#endif
+extern int _Xunlink __P_((char *, int));
+#define	Xunlink(p)	_Xunlink(p, 0)
+extern int _Xrename __P_((char *, char *, int));
+#define	Xrename(f, t)	_Xrename(f, t, 0)
+extern int _Xopen __P_((char *, int, int, int));
+#define	Xopen(p, f, m)	_Xopen(p, f, m, 0)
 #ifdef	_NODOSDRIVE
 #define	Xclose		close
 #define	Xread		read
@@ -141,54 +143,12 @@ extern off_t Xlseek __P_((int, off_t, int));
 extern int Xdup __P_((int));
 extern int Xdup2 __P_((int, int));
 #endif
-#if	MSDOS && defined (_NOUSELFN)
-# ifdef	DJGPP
-#define	_Xmkdir(p, m)	(mkdir(p, m) ? -1 : 0)
-# else
-extern int _Xmkdir __P_((char *, int));
-# endif
-#define	_Xrmdir(p)	(rmdir(p) ? -1 : 0)
-#else
-# if	!MSDOS && defined (_NODOSDRIVE) && defined (_NOKANJIFCONV)
-#define	_Xmkdir		mkdir
-#define	_Xrmdir		rmdir
-# else
-extern int _Xmkdir __P_((char *, int));
-extern int _Xrmdir __P_((char *));
-# endif
-#endif
-#ifdef	_NOROCKRIDGE
-# if	MSDOS && defined (_NOUSELFN)
-#  ifdef	DJGPP
-#define	Xmkdir(p, m)	(mkdir(p, m) ? -1 : 0)
-#  else
-#define	Xmkdir(p, m)	(mkdir(p) ? -1 : 0)
-#  endif
-#define	Xrmdir(p)	(rmdir(p) ? -1 : 0)
-# else
-#  if	!MSDOS && defined (_NODOSDRIVE) && defined (_NOKANJIFCONV)
-#define	Xmkdir		mkdir
-#define	Xrmdir		rmdir
-#  else
-#define	Xmkdir		_Xmkdir
-#define	Xrmdir		_Xrmdir
-#  endif
-# endif
-#else
-extern int Xmkdir __P_((char *, int));
-extern int Xrmdir __P_((char *));
-#endif
-#if	(MSDOS && defined (_NOUSELFN)) \
-|| (!MSDOS && defined (_NODOSDRIVE) && defined (_NOKANJIFCONV))
-#define	_Xfopen		fopen
-#else
-extern FILE *_Xfopen __P_((char *, char *));
-#endif
-#ifdef	_NOROCKRIDGE
-#define	Xfopen		_Xfopen
-#else
-extern FILE *Xfopen __P_((char *, char *));
-#endif
+extern int _Xmkdir __P_((char *, int, int, int));
+#define	Xmkdir(p, m)	_Xmkdir(p, m, 0, 0)
+extern int _Xrmdir __P_((char *, int, int));
+#define	Xrmdir(p)	_Xrmdir(p, 0, 0)
+extern FILE *_Xfopen __P_((char *, char *, int));
+#define	Xfopen(p, t)	_Xfopen(p, t, 0)
 #ifdef	_NODOSDRIVE
 #define	Xfdopen		fdopen
 #define	Xfclose		fclose
@@ -223,7 +183,6 @@ extern int Xpclose __P_((FILE *));
 #endif	/* _NOORIGSHELL */
 
 /* libc.c */
-extern int rename2 __P_((char *, char *));
 extern int stat2 __P_((char *, struct stat *));
 extern char *realpath2 __P_((char *, char *, int));
 extern int _chdir2 __P_((char *));
@@ -238,29 +197,31 @@ extern char *strrchr2 __P_((char *, int));
 extern char *strcpy2 __P_((char *, char *));
 extern char *strncpy2 __P_((char *, char *, int));
 extern int strncpy3 __P_((char *, char *, int *, int));
-extern char *strstr2 __P_((char *, char *));
 extern int strlen2 __P_((char *));
 extern int strlen3 __P_((char *));
 extern int atoi2 __P_((char *));
 extern int putenv2 __P_((char *));
 extern int setenv2 __P_((char *, char *));
 extern char *getenv2 __P_((char *));
-#ifdef	DEBUG
-extern VOID freeenv __P_((VOID_A));
-#endif
 extern int system2 __P_((char *, int));
 extern char *getwd2 __P_((VOID_A));
+extern time_t time2 __P_((VOID_A));
 extern time_t timelocal2 __P_((struct tm *));
 extern char *fgets2 __P_((FILE *, int));
 
 /* file.c */
+#if	MSDOS || defined (_NODOSDRIVE)
+#define	nodospath(p, f)	(f)
+#else
+char *nodospath __P_((char *, char *));
+#endif
+#define	fnodospath(p, i)	nodospath(p, filelist[i].name)
 #if	MSDOS
 extern int logical_access __P_((u_short));
 #else
 extern int logical_access __P_((u_short, uid_t, gid_t));
 #endif
 extern int getstatus __P_((namelist *));
-extern int isdotdir __P_((char *));
 extern int cmplist __P_((CONST VOID_P, CONST VOID_P));
 #ifndef	_NOTREE
 extern int cmptree __P_((CONST VOID_P, CONST VOID_P));
@@ -273,17 +234,18 @@ extern VOID lostcwd __P_((char *));
 extern int safewrite __P_((int, char *, int));
 extern int safecpfile __P_((char *, char *, struct stat *, struct stat *));
 extern int safemvfile __P_((char *, char *, struct stat *, struct stat *));
+extern char *genrandname __P_((char *, int));
 extern int mktmpdir __P_((char *));
 extern int rmtmpdir __P_((char *));
-extern int mktmpfile __P_((char *, char *));
+extern int mktmpfile __P_((char *));
 extern int rmtmpfile __P_((char *));
 extern VOID removetmp __P_((char *, char *, char *));
 extern int forcecleandir __P_((char *, char *));
 #ifndef	_NODOSDRIVE
+extern char *dostmpdir __P_((int));
 extern int tmpdosdupl __P_((char *, char **, int));
 extern int tmpdosrestore __P_((int, char *));
 #endif
-int isexist __P_((char *));
 #ifndef	_NOWRITEFS
 extern VOID arrangedir __P_((int));
 #endif
@@ -293,13 +255,14 @@ extern int rmvfile __P_((char *));
 extern int rmvdir __P_((char *));
 extern int findfile __P_((char *));
 extern int finddir __P_((char *));
-extern int inputattr __P_((namelist *, u_short));
+extern int inputattr __P_((namelist *, int));
 extern int setattr __P_((char *));
 extern int applyfile __P_((int (*)__P_((char *)), char *));
 extern int applydir __P_((char *, int (*)__P_((char *)),
 		int (*)__P_((char *)), int (*)__P_((char *)), int, char *));
 extern int copyfile __P_((char *, int));
 extern int movefile __P_((char *, int));
+extern int forcemovefile __P_((char *));
 
 /* parse.c */
 extern char *skipspace __P_((char *));
@@ -321,8 +284,7 @@ extern char *killmeta __P_((char *));
 #if	!MSDOS && defined (_NOORIGSHELL)
 extern VOID adjustpath __P_((VOID_A));
 #endif
-extern char *includepath __P_((char *, char *, char *));
-extern VOID freevar __P_((char **));
+extern char *includepath __P_((char *, char *, char **));
 #if	(FD < 2) && !defined (_NOARCHIVE)
 extern char *getrange __P_((char *, u_char *, u_char *, u_char *));
 #endif
@@ -339,13 +301,21 @@ extern char *encodestr __P_((char *, int));
 #endif
 
 /* builtin.c */
+#ifndef	_NOARCHIVE
 extern VOID printlaunchcomm __P_((int, int, FILE *));
 extern VOID printarchcomm __P_((int, FILE *));
+# ifndef	_NOBROWSE
+extern VOID freebrowse __P_((launchtable *));
+# endif
+#endif
 extern int freemacro __P_((int));
 extern VOID printmacro __P_((int, FILE *));
+#if	!MSDOS && !defined (_NODOSDRIVE)
+extern int searchdrv __P_((devinfo *, int, char *, int, int, int, int));
 extern int deletedrv __P_((int));
 extern int insertdrv __P_((int, int, char *, int, int, int));
 extern VOID printsetdrv __P_((int, int, FILE *));
+#endif
 #if	!MSDOS && !defined (_NOKEYMAP)
 extern VOID printkeymap __P_((int, char *, int, FILE *));
 #endif
@@ -372,6 +342,7 @@ extern VOID freedefine __P_((VOID_A));
 #define	BL_PENV		"printenv"
 #define	BL_LAUNCH	"launch"
 #define	BL_ARCH		"arch"
+#define	BL_BROWSE	"browse"
 #define	BL_PLAUNCH	"printlaunch"
 #define	BL_PARCH	"printarch"
 #define	BL_BIND		"bind"
@@ -397,8 +368,9 @@ extern VOID freedefine __P_((VOID_A));
 
 /* shell.c */
 extern char *evalcommand __P_((char *, char *, macrostat *, int));
-extern char *inputshellstr __P_((char *, int, int, char *));
+extern char *inputshellstr __P_((char *, int, char *));
 extern char *inputshellloop __P_((int, char *));
+extern int isinternalcomm __P_((char *));
 extern int execmacro __P_((char *, char *, int, int, int));
 #ifdef	_NOORIGSHELL
 extern int execusercomm __P_((char *, char *, int, int, int));
@@ -461,7 +433,8 @@ extern int selectstr __P_((int *, int, int, char *[], int []));
 extern VOID help __P_((int));
 extern int writablefs __P_((char *));
 extern long getblocksize __P_((char *));
-extern VOID getinfofs __P_((char *, long *, long *));
+extern long calcKB __P_((long, long));
+extern int getinfofs __P_((char *, long *, long *, long *));
 extern int infofs __P_((char *));
 
 /* rockridge.c */
@@ -469,6 +442,9 @@ extern int infofs __P_((char *));
 extern int transfilelist __P_((VOID_A));
 extern char *transfile __P_((char *, char *));
 extern char *detransfile __P_((char *, char *, int));
+#else
+#define	transfile(p, b)		(p)
+#define	detransfile(p, b, l)	(p)
 #endif
 
 /* archive.c */
@@ -476,8 +452,12 @@ extern char *detransfile __P_((char *, char *, int));
 extern VOID poparchdupl __P_((VOID_A));
 extern VOID archbar __P_((char *, char *));
 extern VOID copyarcf __P_((reg_t *, char *));
-extern int archchdir __P_((char *));
+extern char *archchdir __P_((char *));
+# ifndef	_NOCOMPLETE
+extern int completearch __P_((char *, int, int, char ***));
+# endif
 extern int launcher __P_((VOID_A));
+extern int dolaunch __P_((launchtable *, int));
 extern int pack __P_((char *));
 extern int unpack __P_((char *, char *, char *, int, int));
 extern char *tmpunpack __P_((int));
@@ -497,11 +477,24 @@ extern VOID evalenv __P_((VOID_A));
 extern VOID freeenvpath __P_((VOID_A));
 #endif
 #ifndef	_NOCUSTOMIZE
+VOID freestrarray __P_((char **, int));
 char **copystrarray __P_((char **, char **, int *, int));
 extern bindtable *copybind __P_((bindtable *, bindtable *));
 # if	!MSDOS && !defined (_NOKEYMAP)
-extern keymaptable *copykeymap __P_((keymaptable *));
 extern VOID freekeymap __P_((keymaptable *));
+extern keymaptable *copykeymap __P_((keymaptable *));
+# endif
+# ifndef	_NOARCHIVE
+extern VOID freelaunch __P_((launchtable *, int));
+extern launchtable *copylaunch __P_((launchtable *, launchtable *,
+	int *, int));
+extern VOID freearch __P_((archivetable *, int));
+extern archivetable *copyarch __P_((archivetable *, archivetable *,
+	int *, int));
+# endif
+# if	!MSDOS && !defined (_NODOSDRIVE)
+extern VOID freedrive __P_((devinfo *));
+extern devinfo *copydrive __P_((devinfo *, devinfo *));
 # endif
 extern VOID rewritecust __P_((int));
 extern int customize __P_((VOID_A));
@@ -511,9 +504,9 @@ extern int customize __P_((VOID_A));
 
 /* browse.c */
 extern VOID helpbar __P_((VOID_A));
-extern char *putmode __P_((char *, u_short));
+extern int putmode __P_((char *, u_short));
 #ifdef	HAVEFLAGS
-extern char *putflags __P_((char *, u_long));
+extern int putflags __P_((char *, u_long));
 #endif
 extern VOID waitmes __P_((VOID_A));
 extern int calcwidth __P_((VOID_A));
