@@ -32,7 +32,7 @@ extern int errno;
 #include <mntent.h>
 typedef struct mntent	mnt_t;
 #define	getmntent2(fp, mntp)	getmntent(fp)
-#define	hasmntopt2(mntp, opt)	strstr2((mntp) -> mnt_opts, opt)
+#define	hasmntopt2(mntp, opt)	strmntopt((mntp) -> mnt_opts, opt)
 #endif	/* USEMNTENTH */
 
 #ifdef	USEMNTTABH
@@ -41,7 +41,7 @@ typedef struct mntent	mnt_t;
 typedef struct mnttab	mnt_t;
 #define	setmntent		fopen
 #define	getmntent2(fp, mntp)	(getmntent(fp, mntp) ? NULL : mntp)
-#define	hasmntopt2(mntp, opt)	strstr2((mntp) -> mnt_mntopts, opt)
+#define	hasmntopt2(mntp, opt)	strmntopt((mntp) -> mnt_mntopts, opt)
 #define	endmntent		fclose
 #define	mnt_dir		mnt_mountp
 #define	mnt_fsname	mnt_special
@@ -77,7 +77,7 @@ typedef struct _mnt_t {
 } mnt_t;
 static FILE *NEAR setmntent __P_((char *, char *));
 static mnt_t *NEAR getmntent2 __P_((FILE *, mnt_t *));
-#define	hasmntopt2(mntp, opt)	strstr2((mntp) -> mnt_opts, opt)
+#define	hasmntopt2(mntp, opt)	strmntopt((mntp) -> mnt_opts, opt)
 #if	defined (USEMNTINFO) || defined (USEGETMNT)
 #define	endmntent(fp)
 #else
@@ -92,7 +92,7 @@ static int mnt_size = 0;
 typedef struct fstab	mnt_t;
 #define	setmntent(file, mode)	(FILE *)(setfsent(), NULL)
 #define	getmntent2(fp, mntp)	getfsent()
-#define	hasmntopt2(mntp, opt)	strstr2((mntp) -> fs_mntops, opt)
+#define	hasmntopt2(mntp, opt)	strmntopt((mntp) -> fs_mntops, opt)
 #define	endmntent(fp)		endfsent()
 #define	mnt_dir		fs_file
 #define	mnt_fsname	fs_spec
@@ -106,7 +106,7 @@ typedef struct _mnt_t {
 	char *mnt_type;
 	char *mnt_opts;
 } mnt_t;
-#define	hasmntopt2(mntp, opt)	strstr2((mntp) -> mnt_opts, opt)
+#define	hasmntopt2(mntp, opt)	strmntopt((mntp) -> mnt_opts, opt)
 # ifdef	PC98
 # define	PT_FAT12	0x81	/* 0x80 | 0x01 */
 # define	PT_FAT16	0x91	/* 0x80 | 0x11 */
@@ -155,6 +155,9 @@ typedef struct fs_data	statfs_t;
 #define	blocksize(fs)		1024
 #define	f_bsize		fd_req.bsize
 #define	f_files		fd_req.gtot
+#define	f_blocks	fd_req.btot
+#define	f_bfree		fd_req.bfree
+#define	f_bavail	fd_req.bfreen
 #endif	/* USEFSDATA */
 
 #ifdef	USEFFSIZE
@@ -211,7 +214,6 @@ extern int dosstatfs __P_((int, char *));
 #endif
 extern char *malloc2 __P_((ALLOC_T));
 extern char *realloc2 __P_((VOID_P, ALLOC_T));
-extern char *strstr2 __P_((char *, char *));
 extern char *ascnumeric __P_((char *, long, int, int));
 extern int kanjiputs __P_((char *));
 extern int kanjiputs2 __P_((char *, int, int));
@@ -243,6 +245,9 @@ extern int needbavail;
 #ifndef	MNTTYPE_FFS
 #define	MNTTYPE_FFS	"ffs"	/* NetBSD, OpenBSD */
 #endif
+#ifndef	MNTTYPE_HFS
+#define	MNTTYPE_HFS	"hfs"	/* Darwin, HP-UX */
+#endif
 #ifndef	MNTTYPE_EXT2
 #define	MNTTYPE_EXT2	"ext2"	/* Linux */
 #endif
@@ -281,27 +286,27 @@ static int NEAR code2str __P_((char *, int));
 static int NEAR checkline __P_((int));
 VOID help __P_((int));
 static int NEAR getfsinfo __P_((char *, statfs_t *, mnt_t *));
+static char *NEAR strmntopt __P_((char *, char *));
 int writablefs __P_((char *));
 long getblocksize __P_((char *));
 static int NEAR info1line __P_((int, char *, long, char *, char *));
-#if	!defined (USEFSDATA) || !defined (_NODOSDRIVE)
-static long NEAR calcKB __P_((long, long));
-#endif
-VOID getinfofs __P_((char *, long *, long *));
+long calcKB __P_((long, long));
+int getinfofs __P_((char *, long *, long *, long *));
 int infofs __P_((char *));
 
 static int keycodelist[] = {
 	K_HOME, K_END, K_DL, K_IL, K_DC, K_IC,
-	K_BEG, K_EOL, K_NPAGE, K_PPAGE, K_CLR, K_ENTER,
-	K_BS, '\t', CR, ESC
+	K_BEG, K_EOL, K_NPAGE, K_PPAGE, K_CLR, K_ENTER, K_HELP,
+	K_BS, '\t', K_CR, K_ESC
 };
 #define	KEYCODESIZ	((int)(sizeof(keycodelist) / sizeof(int)))
 static char *keystrlist[] = {
 	"Home", "End", "DelLin", "InsLin", "Del", "Ins",
-	"Beg", "Eol", "PageDn", "PageUp", "Clr", "Enter",
+	"Beg", "Eol", "PageDn", "PageUp", "Clr", "Enter", "Help",
 	"Bs", "Tab", "Ret", "Esc"
 };
 
+
 static int NEAR code2str(buf, code)
 char *buf;
 int code;
@@ -326,7 +331,8 @@ int code;
 #endif
 		else if (isctl(code))
 			sprintf(buf, "Ctrl-%c ", (code + '@') & 0x7f);
-		else if (!ismsb(code)) sprintf(buf, "'%c'    ", code);
+		else if (code < K_MIN && !ismsb(code))
+			sprintf(buf, "'%c'    ", code);
 		else return(0);
 	}
 	return(1);
@@ -349,8 +355,8 @@ int mode;
 	int i, j, c, x, y;
 
 	if (distributor && *distributor) {
-		i = n_column - strlen(distributor) - 24;
-		locate(i, LHELP);
+		i = n_column - (int)strlen(distributor) - 24;
+		locate(i, L_HELP);
 		putch2('[');
 		putterm(t_standout);
 		cprintf2(" Distributed by: %s ", distributor);
@@ -424,7 +430,6 @@ mnt_t *mntp;
 	static char *fsname = NULL;
 	static char *dir = NULL;
 	static char *type = NULL;
-	static char *opts = NULL;
 	struct vfs_ent *entp;
 	struct vmount *vmntp;
 	char *cp, *buf, *host;
@@ -468,7 +473,7 @@ mnt_t *mntp;
 	mntp -> mnt_fsname = fsname;
 	mntp -> mnt_dir = dir;
 	mntp -> mnt_type = (type) ? type : "???";
-	mntp -> mnt_opts = (opts) ? opts : "";
+	mntp -> mnt_opts = (vmntp -> vmt_flags & MNT_READONLY) ? "ro" : "";
 
 	mnt_ptr += vmntp -> vmt_length;
 	return(mntp);
@@ -674,7 +679,10 @@ mnt_t *mntbuf;
 		memset((char *)fsbuf, 0xff, sizeof(statfs_t));
 	}
 #else	/* !MSDOS */
-	mnt_t *mntp, mnt;
+# if	!defined (USEMNTENTH) && !defined (USEGETFSENT)
+	mnt_t mnt;
+# endif
+	mnt_t *mntp;
 	FILE *fp;
 	char *dir, fsname[MAXPATHLEN];
 	int len, match;
@@ -702,20 +710,20 @@ mnt_t *mntbuf;
 		if (buf[sizeof(long) * 3] & 001)
 			mntbuf -> mnt_type = MNTTYPE_FAT32;
 		fsbuf -> f_bsize = *((long *)&(buf[sizeof(long) * 0]));
-#ifdef	USEFSDATA
+# ifdef	USEFSDATA
 		fsbuf -> fd_req.btot = calcKB(fsbuf -> f_bsize,
 			*((long *)&(buf[sizeof(long) * 1])));
 		fsbuf -> fd_req.bfree =
 		fsbuf -> fd_req.bfreen = calcKB(fsbuf -> f_bsize,
 			*((long *)&(buf[sizeof(long) * 2])));
-#else	/* !USEFSDATA */
-#ifdef	USESTATVFSH
+# else	/* !USEFSDATA */
+#  ifdef	USESTATVFSH
 		fsbuf -> f_frsize = 0;
-#endif
+#  endif
 		fsbuf -> f_blocks = *((long *)&(buf[sizeof(long) * 1]));
 		fsbuf -> f_bfree =
 		fsbuf -> f_bavail = *((long *)&(buf[sizeof(long) * 2]));
-#endif	/* !USEFSDATA */
+# endif	/* !USEFSDATA */
 		fsbuf -> f_files = -1;
 		return(0);
 	}
@@ -774,6 +782,21 @@ mnt_t *mntbuf;
 	return(0);
 }
 
+static char *NEAR strmntopt(s1, s2)
+char *s1, *s2;
+{
+	char *cp;
+	int len;
+
+	len = strlen(s2);
+	for (cp = s1; cp && *cp;) {
+		if (!strncmp(cp, s2, len) && (!cp[len] || cp[len] == ','))
+			return(cp);
+		if ((cp = strchr(cp, ','))) cp++;
+	}
+	return(NULL);
+}
+
 int writablefs(path)
 char *path;
 {
@@ -802,6 +825,11 @@ char *path;
 	else if (!strcmp(mntbuf.mnt_type, MNTTYPE_SYSV)
 	|| !strcmp(mntbuf.mnt_type, MNTTYPE_DGUX)) return(3);
 	else
+#endif
+#ifdef	DARWIN
+	/* Macintosh HFS+ is pseudo file system covered with skin */
+	if (!strcmp(mntbuf.mnt_type, MNTTYPE_HFS)) return(0);
+	else
 #endif
 	if (!strcmp(mntbuf.mnt_type, MNTTYPE_PC)) return(4);
 	else if (!strcmp(mntbuf.mnt_type, MNTTYPE_DOS7)) return(5);
@@ -861,12 +889,12 @@ char *s, *unit;
 	return(checkline(++y));
 }
 
-#if	!defined (USEFSDATA) || !defined (_NODOSDRIVE)
-static long NEAR calcKB(block, byte)
+long calcKB(block, byte)
 long block, byte;
 {
-	if (block < 0 || byte <= 0) return(-1);
-	if (byte >= 1024) {
+	if (block < 0 || byte <= 0) return(-1L);
+	if (byte == 1024) return(block);
+	else if (byte > 1024) {
 		byte = (byte + 512) / 1024;
 		return(block * byte);
 	}
@@ -875,11 +903,10 @@ long block, byte;
 		return(block / byte);
 	}
 }
-#endif
 
-VOID getinfofs(path, totalp, freep)
+int getinfofs(path, totalp, freep, bsizep)
 char *path;
-long *totalp, *freep;
+long *totalp, *freep, *bsizep;
 {
 	statfs_t fsbuf;
 	mnt_t mntbuf;
@@ -887,20 +914,14 @@ long *totalp, *freep;
 #ifndef	_NODOSDRIVE
 	needbavail = 1;
 #endif
-	if (getfsinfo(path, &fsbuf, &mntbuf) < 0)
-		*totalp = *freep = -1;
-	else {
-#ifdef	USEFSDATA
-		*totalp = fsbuf.fd_req.btot;
-		*freep = fsbuf.fd_req.bfreen;
-#else
-		*totalp = calcKB(fsbuf.f_blocks, blocksize(fsbuf));
-		*freep = calcKB(fsbuf.f_bavail, blocksize(fsbuf));
-#endif
-	}
+	if (getfsinfo(path, &fsbuf, &mntbuf) < 0) return(-1);
+	*totalp = fsbuf.f_blocks;
+	*freep = fsbuf.f_bavail;
+	*bsizep = blocksize(fsbuf);
 #ifndef	_NODOSDRIVE
 	needbavail = 0;
 #endif
+	return(0);
 }
 
 int infofs(path)
@@ -927,12 +948,6 @@ char *path;
 	y = info1line(y, FSNAM_K, 0L, mntbuf.mnt_fsname, NULL);
 	y = info1line(y, FSMNT_K, 0L, mntbuf.mnt_dir, NULL);
 	y = info1line(y, FSTYP_K, 0L, mntbuf.mnt_type, NULL);
-#ifdef	USEFSDATA
-	y = info1line(y, FSTTL_K, fsbuf.fd_req.btot, NULL, "Kbytes");
-	y = info1line(y, FSUSE_K,
-		fsbuf.fd_req.btot - fsbuf.fd_req.bfree, NULL, "Kbytes");
-	y = info1line(y, FSAVL_K, fsbuf.fd_req.bfreen, NULL, "Kbytes");
-#else
 	y = info1line(y, FSTTL_K,
 		calcKB(fsbuf.f_blocks, blocksize(fsbuf)), NULL, "Kbytes");
 	y = info1line(y, FSUSE_K,
@@ -940,7 +955,6 @@ char *path;
 		NULL, "Kbytes");
 	y = info1line(y, FSAVL_K,
 		calcKB(fsbuf.f_bavail, blocksize(fsbuf)), NULL, "Kbytes");
-#endif
 	y = info1line(y, FSBSZ_K, fsbuf.f_bsize, NULL, "bytes");
 	y = info1line(y, FSINO_K, fsbuf.f_files, NULL, UNIT_K);
 	if (y > LFILETOP + 1) {
