@@ -11,17 +11,15 @@
 #include "kctype.h"
 #include "kanji.h"
 
-#ifndef	_NOORIGSHELL
-#include "system.h"
-#endif
-
 #ifdef	USESTDARGH
 #include <stdarg.h>
 #else
 #include <varargs.h>
 #endif
 
-#define	LIMITSELECTWARN	100
+#ifndef	_NOORIGSHELL
+#include "system.h"
+#endif
 
 extern char **history[];
 extern short histsize[];
@@ -29,15 +27,19 @@ extern int curcolumns;
 extern int minfilename;
 extern int sizeinfo;
 extern int hideclock;
+#ifdef	SIGALRM
+extern int clockmode;
+#endif
 extern char *promptstr;
 #ifndef	_NOORIGSHELL
 extern char *promptstr2;
-extern int shellmode;
 #endif
 #ifndef	DECLERRLIST
 extern char *sys_errlist[];
 #endif
 
+#define	LIMITSELECTWARN	100
+
 static char *NEAR trquote __P_((char *, int, int *));
 static int NEAR vlen __P_((char *, int));
 static int NEAR rlen __P_((char *, int));
@@ -46,6 +48,9 @@ static VOID NEAR rightcursor __P_((VOID_A));
 static VOID NEAR leftcursor __P_((VOID_A));
 static VOID NEAR upcursor __P_((VOID_A));
 static VOID NEAR downcursor __P_((VOID_A));
+#ifndef	_NOORIGSHELL
+static VOID NEAR movecursor __P_((char *, char *, int));
+#endif
 static VOID NEAR Xlocate __P_((int, int));
 static VOID NEAR setcursor __P_((int, int, int, int));
 static int NEAR rightchar __P_((char *, int *, int, int, int, int, int));
@@ -100,24 +105,24 @@ static int emulatekey[] = {
 	K_PPAGE, K_NPAGE, K_ENTER, K_ESC
 };
 static char emacskey[] = {
-	CTRL('P'), CTRL('N'), CTRL('F'), CTRL('B'),
-	K_ESC, CTRL('D'), CTRL('Q'), CTRL('K'),
-	K_ESC, K_ESC, CTRL('A'), CTRL('E'),
-	CTRL('V'), CTRL('Y'), CTRL('O'), CTRL('G')
+	K_CTRL('P'), K_CTRL('N'), K_CTRL('F'), K_CTRL('B'),
+	K_ESC, K_CTRL('D'), K_CTRL('Q'), K_CTRL('K'),
+	K_ESC, K_ESC, K_CTRL('A'), K_CTRL('E'),
+	K_CTRL('V'), K_CTRL('Y'), K_CTRL('O'), K_CTRL('G')
 };
 #define	EMACSKEYSIZ	((int)(sizeof(emacskey) / sizeof(char)))
 static char vikey[] = {
 	'k', 'j', 'l', 'h',
 	K_ESC, 'x', K_ESC, 'D',
 	'g', 'G', '0', '$',
-	CTRL('B'), CTRL('F'), 'o', K_ESC
+	K_CTRL('B'), K_CTRL('F'), 'o', K_ESC
 };
 #define	VIKEYSIZ	((int)(sizeof(vikey) / sizeof(char)))
 static char wordstarkey[] = {
-	CTRL('E'), CTRL('X'), CTRL('D'), CTRL('S'),
-	CTRL('V'), CTRL('G'), CTRL(']'), CTRL('Y'),
-	CTRL('W'), CTRL('Z'), CTRL('A'), CTRL('F'),
-	CTRL('R'), CTRL('C'), CTRL('N'), K_ESC
+	K_CTRL('E'), K_CTRL('X'), K_CTRL('D'), K_CTRL('S'),
+	K_CTRL('V'), K_CTRL('G'), K_CTRL(']'), K_CTRL('Y'),
+	K_CTRL('W'), K_CTRL('Z'), K_CTRL('A'), K_CTRL('F'),
+	K_CTRL('R'), K_CTRL('C'), K_CTRL('N'), K_ESC
 };
 #define	WORDSTARKEYSIZ	((int)(sizeof(wordstarkey) / sizeof(char)))
 #endif
@@ -155,6 +160,11 @@ int sig, eof;
 	}
 #endif
 
+#ifdef	SIGALRM
+	sig = (clockmode) ? SIGALRM : 0;
+#else
+	sig = 0;
+#endif
 	ch = getkey2(sig);
 	if (eof && ch == cc_eof) return(-1);
 
@@ -169,7 +179,7 @@ int sig, eof;
 	else if (!strcmp(editmode, "vi")) do {
 		vimode |= 1;
 		if (vimode & 2) switch (ch) {
-			case CTRL('V'):
+			case K_CTRL('V'):
 				if (vimode & 4) vimode = 1;
 				ch = K_IL;
 				break;
@@ -394,9 +404,37 @@ static VOID NEAR downcursor(VOID_A)
 	putterm(c_down);
 }
 
+#ifndef	_NOORIGSHELL
+static VOID NEAR movecursor(s, s2, n)
+char *s, *s2;
+int n;
+{
+	char *cp;
+
+	if (!*s || !(cp = tparamstr(s, n, n))) while (n--) putterm(s2);
+	else {
+		putterm(cp);
+		free(cp);
+	}
+}
+#endif	/* !_NOORIGSHELL */
+
 static VOID NEAR Xlocate(x, y)
 int x, y;
 {
+#ifndef	_NOORIGSHELL
+	if (shellmode) {
+		x += xpos;
+		if (x < win_x) movecursor(c_nleft, c_left, win_x - x);
+		else if (x > win_x) movecursor(c_nright, c_right, x - win_x);
+		if (y < win_y) movecursor(c_nup, c_up, win_y - y);
+		else if (y > win_y) movecursor("", c_scrollforw, y - win_y);
+
+		win_x = x;
+		win_y = y;
+		return;
+	}
+#endif	/* !_NOORIGSHELL */
 	while (ypos + y >= n_line) {
 		locate(0, n_line - 1);
 		putterm(c_scrollforw);
@@ -491,7 +529,7 @@ char *s;
 int cx, len, plen, max, linemax, ins;
 {
 	char *dupl;
-	int dx, dy, i, j, l, f1, ptr;
+	int dx, dy, i, j, l, f1, ptr, nlen;
 #if	!MSDOS
 	int f2;
 #endif
@@ -499,7 +537,7 @@ int cx, len, plen, max, linemax, ins;
 	dupl = trquote(&(s[cx]), len - cx, &l);
 	cx = vlen(s, cx);
 	len = cx + l;
-	len += ins;
+	nlen = len + ins;
 	dy = (cx + plen) / linemax;
 	i = (dy + 1) * linemax - plen;
 	j = linemax - (cx + plen) % linemax;
@@ -507,9 +545,16 @@ int cx, len, plen, max, linemax, ins;
 
 #if	!MSDOS
 	if (*c_insert) {
+# ifndef	_NOORIGSHELL
+		if (shellmode);
+		else
+# endif
 		if (ypos + dy >= n_line - 1
-		&& xpos + (cx + plen) % linemax + len - ins - cx + j
-		>= n_column) {
+		&& xpos + (cx + plen) % linemax + (len - cx) + j >= n_column) {
+			/*
+			 * In the case that current line is the last line,
+			 * and the end of string will reach over the last line.
+			 */
 			while (ypos + dy >= n_line - 1) {
 				locate(0, n_line - 1);
 				putterm(c_scrollforw);
@@ -519,25 +564,33 @@ int cx, len, plen, max, linemax, ins;
 		}
 		while (j--) putterm(c_insert);
 
-		if (i < len) {
-			while (i < len) {
+		if (i < nlen) {
+			while (i < nlen) {
 				ptr = i - 1 - cx;
+				/*
+				 * Whether if the end of line is kanji 1st byte
+				 * f1: after
+				 * f2: before
+				 */
 				if (ptr < ins) f1 = 0;
 				else f1 = (onkanji1(dupl, ptr - ins)) ? 1 : 0;
 				f2 = (onkanji1(dupl, ptr)) ? 1 : 0;
 				if (xpos + linemax < n_column) {
 					Xlocate(linemax, dy);
-					j = n_column - xpos - linemax;
+					j = n_column - (xpos + linemax);
 					if (j > ins) j = ins;
-					if (f2) j++;
 					if (f1) {
 						rightcursor();
 						j--;
 					}
 					putcursor(' ', j);
 				}
+# ifndef	_NOORIGSHELL
+				if (shellmode) dy++;
+				else
+# endif
 				if (ypos + ++dy >= n_line - 1
-				&& xpos + len - i >= n_column) {
+				&& xpos + nlen - i >= n_column) {
 					while (ypos + dy >= n_line - 1) {
 						locate(0, n_line - 1);
 						putterm(c_scrollforw);
@@ -558,8 +611,8 @@ int cx, len, plen, max, linemax, ins;
 					if (f1) dupl[rlen(dupl, ptr)] = ' ';
 					if (f2) l++;
 				}
-				if (ptr + l > len - cx) l = len - cx - ptr;
-				kanjiputs2(dupl, l, ptr);
+				if (ptr + l > nlen - cx) l = nlen - cx - ptr;
+				win_x += kanjiputs2(dupl, l, ptr);
 				i += linemax;
 			}
 			setcursor(cx, plen, max, linemax);
@@ -573,10 +626,18 @@ int cx, len, plen, max, linemax, ins;
 		l = i - cx - ins;
 		dx = (cx + ins + plen) % linemax;
 
-		while (i < len) {
+		while (i < nlen) {
 			ptr = i - 1 - cx;
+			/*
+			 * Whether if the end of line is kanji 1st byte
+			 * f1: after
+			 */
 			if (ptr < ins) f1 = l = 0;
 			else f1 = (onkanji1(dupl, ptr - ins)) ? 1 : 0;
+# ifndef	_NOORIGSHELL
+			if (shellmode);
+			else
+# endif
 			if (ypos + dy >= n_line - 1) {
 				while (ypos + dy >= n_line - 1) {
 					locate(0, n_line - 1);
@@ -585,13 +646,13 @@ int cx, len, plen, max, linemax, ins;
 				}
 				Xlocate(dx, dy);
 			}
-			if ((l += f1) > 0) kanjiputs2(dupl, l, j);
-			if (!f1) putch2(' ');
+			if ((l += f1) > 0) win_x += kanjiputs2(dupl, l, j);
+			if (!f1) putcursor(' ', 1);
 
 			Xlocate(0, ++dy);
 			l = linemax;
 			if (ptr < ins - 1) {
-				putch2(' ');
+				putcursor(' ', 1);
 				ptr++;
 				Xlocate(ins - ptr, dy);
 				l -= ins - ptr;
@@ -601,17 +662,27 @@ int cx, len, plen, max, linemax, ins;
 				ptr -= ins - 1;
 				if (f1) dupl[rlen(dupl, ptr)] = ' ';
 			}
-			if (ptr + l > len - ins - cx) l = len - ins - cx - ptr;
+			if (ptr + l > len - cx) l = len - cx - ptr;
 			j = ptr;
 			i += linemax;
 			dx = 0;
 		}
 
-		l = len - ins - cx - j;
-		if (l > 0) kanjiputs2(dupl, l, j);
+		l = len - cx - j;
+		if (l > 0) win_x += kanjiputs2(dupl, l, j);
 		setcursor(cx, plen, max, linemax);
 	}
-	if (i == len && len < max && ypos + dy >= n_line - 1) {
+# ifndef	_NOORIGSHELL
+	if (shellmode) {
+		if (((nlen + plen) % linemax) == 1) {
+			setcursor(nlen, plen, max, linemax);
+			putterm(l_clear);
+			setcursor(cx, plen, max, linemax);
+		}
+	}
+	else
+# endif
+	if (i == nlen && nlen < max && ypos + dy >= n_line - 1) {
 		while (ypos + dy >= n_line - 1) {
 			locate(0, n_line - 1);
 			putterm(c_scrollforw);
@@ -628,7 +699,7 @@ char *s;
 int cx, len, plen, max, linemax, del;
 {
 	char *dupl;
-	int dy, i, j, l, f1, ptr;
+	int dy, i, j, l, f1, ptr, nlen;
 #if	!MSDOS
 	int f2;
 #endif
@@ -636,7 +707,7 @@ int cx, len, plen, max, linemax, del;
 	dupl = trquote(&(s[cx]), len - cx, &l);
 	cx = vlen(s, cx);
 	len = cx + l;
-	len -= del;
+	nlen = len - del;
 	dy = (cx + plen) / linemax;
 	i = (dy + 1) * linemax - plen;
 
@@ -646,10 +717,14 @@ int cx, len, plen, max, linemax, del;
 		if (j > del) j = del;
 		while (j--) putterm(c_delete);
 
-		if (i < len + del) {
-			while (i < len + del) {
+		if (i < len) {
+			while (i < len) {
 				ptr = i - 1 - cx;
-				if (ptr >= len - cx) f1 = 0;
+				/*
+				 * Whether if the end of line is kanji 1st byte
+				 * f1: after
+				 */
+				if (ptr >= nlen - cx) f1 = 0;
 				else f1 = (onkanji1(dupl, ptr + del)) ? 1 : 0;
 				f2 = (onkanji1(dupl, ptr)) ? 1 : 0;
 				j = linemax - del + f2;
@@ -663,12 +738,12 @@ int cx, len, plen, max, linemax, del;
 					ptr += f2;
 				}
 				Xlocate(j, dy);
-				if (ptr + l > len + del) l = len + del - ptr;
-				if (l > 0) kanjiputs2(dupl, l, ptr);
-				if (!f1) putch2(' ');
+				if (ptr + l > len) l = len - ptr;
+				if (l > 0) win_x += kanjiputs2(dupl, l, ptr);
+				if (!f1) putcursor(' ', 1);
 				Xlocate(0, ++dy);
 				for (j = 0; j < del; j++) putterm(c_delete);
-				if (f1) putch2(' ');
+				if (f1) putcursor(' ', 1);
 				i += linemax;
 			}
 			setcursor(cx, plen, max, linemax);
@@ -680,24 +755,28 @@ int cx, len, plen, max, linemax, del;
 		j = del;
 		l = i - cx;
 
-		while (i < len + del) {
+		while (i < len) {
 			ptr = i - 1 - cx;
-			if (ptr >= len - cx) f1 = 0;
+			/*
+			 * Whether if the end of line is kanji 1st byte
+			 * f1: after
+			 */
+			if (ptr >= nlen - cx) f1 = 0;
 			else f1 = (onkanji1(dupl, ptr + del)) ? 1 : 0;
-			if ((l += f1) > 0) kanjiputs2(dupl, l, j);
-			if (!f1) putch2(' ');
+			if ((l += f1) > 0) win_x += kanjiputs2(dupl, l, j);
+			if (!f1) putcursor(' ', 1);
 
 			Xlocate(0, ++dy);
 			l = linemax + 1;
 			ptr += del + 1;
 			if (f1) dupl[rlen(dupl, ptr)] = ' ';
-			if (ptr + l > len + del - cx) l = len + del - cx - ptr;
+			if (ptr + l > len - cx) l = len - cx - ptr;
 			j = ptr;
 			i += linemax;
 		}
 
-		l = len + del - cx - j;
-		if (l > 0) kanjiputs2(dupl, l, j);
+		l = len - cx - j;
+		if (l > 0) win_x += kanjiputs2(dupl, l, j);
 		putterm(l_clear);
 		setcursor(cx, plen, max, linemax);
 	}
@@ -732,6 +811,14 @@ int cx, len, plen, max, linemax;
 	dy = (cx + plen) / linemax;
 	i = (dy + 1) * linemax - plen;
 	if (i < len) {
+#ifndef	_NOORIGSHELL
+		if (shellmode) while (i < len) {
+			Xlocate(0, ++dy);
+			putterm(l_clear);
+			i += linemax;
+		}
+		else
+#endif
 		while (i < len) {
 			if (ypos + ++dy >= n_line) break;
 			locate(xpos, ypos + dy);
@@ -758,13 +845,13 @@ int cx, len, plen, max, linemax;
 		putterm(l_clear);
 		if (stable_standout) putterm(end_standout);
 		if (onkanji1(dupl, i + width - 1)) {
-			kanjiputs2(dupl, ++width, i);
+			win_x += kanjiputs2(dupl, ++width, i);
 			Xlocate(0, y);
-			putch2(' ');
+			putcursor(' ', 1);
 		}
 		else {
-			kanjiputs2(dupl, width, i);
-			putch2(' ');
+			win_x += kanjiputs2(dupl, width, i);
+			putcursor(' ', 1);
 			Xlocate(0, y);
 		}
 		i += width;
@@ -772,7 +859,11 @@ int cx, len, plen, max, linemax;
 	}
 	putterm(l_clear);
 	if (stable_standout) putterm(end_standout);
-	kanjiputs2(dupl, len - i, i);
+	win_x += kanjiputs2(dupl, len - i, i);
+#ifndef	_NOORIGSHELL
+	if (shellmode);
+	else
+#endif
 	for (; y * linemax - plen < max; y++) {
 		if (ypos + y >= n_line) break;
 		locate(xpos, ypos + y);
@@ -852,52 +943,78 @@ int ins, quote;
 	dupl = malloc2(ins * 2 + 1);
 	insertchar(s, cx, len, plen, max, linemax, vlen(strins, ins));
 	insshift(s, cx, len, ins);
+	for (i = 0; i < ins; i++) s[cx + i] = ' ';
+	len += ins;
 	for (i = j = 0; i < ins; i++, j++) {
 		if (isctl(strins[i])) {
-			if (len + ins + j - i >= max)
+			if (len + j - i >= max)
 				dupl[j] = s[cx + j] = '?';
 			else {
+				insertchar(s, cx, len, plen, max, linemax, 1);
+				insshift(s, cx + j, len, 1);
 				dupl[j] = '^';
 				s[cx + j] = QUOTE;
-				dupl[++j] = (strins[i] + '@') & 0x7f;
+				j++;
+				dupl[j] = (strins[i] + '@') & 0x7f;
 				s[cx + j] = strins[i];
 			}
 		}
 #ifdef	CODEEUC
 		else if (isekana(strins, i)) {
-			if (len + ins + j - i >= max)
+			if (len + j - i >= max)
 				dupl[j] = s[cx + j] = '?';
 			else {
 				dupl[j] = s[cx + j] = strins[i];
-				dupl[j + 1] = s[cx + j + 1] = strins[++i];
 				j++;
+				dupl[j] = s[cx + j] = strins[++i];
 			}
 		}
 #endif
 		else if (iskanji1(strins, i)) {
-			if (len + ins + j - i >= max)
+			if (len + j - i >= max)
 				dupl[j] = s[cx + j] = '?';
 			else {
 				dupl[j] = s[cx + j] = strins[i];
-				dupl[j + 1] = s[cx + j + 1] = strins[++i];
 				j++;
+				dupl[j] = s[cx + j] = strins[++i];
 			}
 		}
 #if	MSDOS && defined (_NOORIGSHELL)
-		else if (strchr(DQ_METACHAR, strins[i])) {
+		else if (strchr(DQ_METACHAR, strins[i])
 #else
-		else if (quote == '"' && strchr(DQ_METACHAR, strins[i])) {
-#endif
-			if (len + ins + j - i >= max)
+		else if ((quote == '\'' && strins[i] == '\'')
+		|| (quote == '"' && strins[i] == '!')) {
+			f = 3;
+			if (!strins[i + 1]) f--;
+			if (len + j - i + f > max)
 				dupl[j] = s[cx + j] = '?';
 			else {
+				insertchar(s, cx, len, plen, max, linemax, f);
+				insshift(s, cx + j, len, f);
+				dupl[j] = s[cx + j] = quote;
+				j++;
+				dupl[j] = s[cx + j] = PMETA;
+				j++;
+				dupl[j] = s[cx + j] = strins[i];
+				j++;
+				dupl[j] = s[cx + j] = quote;
+			}
+		}
+		else if ((quote == '"' && strchr(DQ_METACHAR, strins[i]))
+#endif
+		|| (!quote && strchr(METACHAR, strins[i]))) {
+			if (len + j - i >= max)
+				dupl[j] = s[cx + j] = '?';
+			else {
+				insertchar(s, cx, len, plen, max, linemax, 1);
+				insshift(s, cx + j, len, 1);
 #if	MSDOS && defined (_NOORIGSHELL)
-				dupl[j] = s[cx + j] =
+				dupl[j] = s[cx + j] = strins[i];
 #else
 				dupl[j] = s[cx + j] = PMETA;
 #endif
-				dupl[j + 1] = s[cx + j + 1] = strins[i];
 				j++;
+				dupl[j] = s[cx + j] = strins[i];
 			}
 		}
 		else dupl[j] = s[cx + j] = strins[i];
@@ -911,7 +1028,15 @@ int ins, quote;
 	ptr = 0;
 	cx = (cx + plen) % linemax;
 	while (i < len) {
+		/*
+		 * Whether if the end of line is kanji 1st byte
+		 * f: after
+		 */
 		f = (onkanji1(dupl, i - 1)) ? 1 : 0;
+#ifndef	_NOORIGSHELL
+		if (shellmode);
+		else
+#endif
 		if (ypos + dy >= n_line - 1) {
 			while (ypos + dy >= n_line - 1) {
 				locate(0, n_line - 1);
@@ -920,7 +1045,7 @@ int ins, quote;
 			}
 			Xlocate(cx, dy);
 		}
-		kanjiputs2(dupl, i - ptr + f, ptr);
+		win_x += kanjiputs2(dupl, i - ptr + f, ptr);
 		Xlocate(0, ++dy);
 		if (f) dupl[rlen(dupl, i)] = ' ';
 		ptr = i;
@@ -960,15 +1085,15 @@ char **argv;
 		qsort(selectlist, argc, sizeof(namelist), cmplist);
 		sorton = dupsorton;
 
-		if (lcmdline > 0) {
-			char buf[20 + 1];
+		if (lcmdline < 0) {
+			char buf[MAXLONGWIDTH + 1];
 			int j, n;
 
 			n = 1;
 			cputs2("\r\n");
 			if (argc < LIMITSELECTWARN) i = 'Y';
 			else {
-				ascnumeric(buf, argc, 0, 20);
+				ascnumeric(buf, argc, 0, MAXLONGWIDTH);
 				cputs2("There are ");
 				cputs2(buf);
 				cputs2(" possibilities.  ");
@@ -1005,11 +1130,21 @@ char **argv;
 				}
 			}
 
+# ifndef	_NOORIGSHELL
+			if (shellmode) {
+				win_x = xpos;
+				win_y = 0;
+			}
+			else
+# endif
 			if (ypos + n < n_line - 1) ypos += n;
 			else ypos = n_line - 1;
 			for (i = 0; i < argc; i++) free(selectlist[i].name);
 			free(selectlist);
 			selectlist = NULL;
+			minfilename = dupminfilename;
+			curcolumns = dupcolumns;
+
 			return;
 		}
 
@@ -1049,7 +1184,7 @@ char *s;
 int cx, len, plen, max, linemax, comline, cont;
 {
 # if	!MSDOS || !defined (_NOORIGSHELL)
-	int bq;
+	int bq, hadmeta;
 # endif
 	char *cp1, *cp2, **argv;
 	int i, l, ins, top, fix, argc, quote, quoted, hasmeta;
@@ -1061,7 +1196,7 @@ int cx, len, plen, max, linemax, comline, cont;
 	}
 
 # if	!MSDOS || !defined (_NOORIGSHELL)
-	bq = 0;
+	bq = hadmeta = 0;
 # endif
 	quote = '\0';
 	quoted = 0;
@@ -1074,6 +1209,13 @@ int cx, len, plen, max, linemax, comline, cont;
 			quote = '\0';
 		}
 		else if (iskanji1(s, i)) i++;
+# if	!MSDOS || !defined (_NOORIGSHELL)
+		else if (quote == '\'');
+		else if (isnmeta(s, i, quote, cx)) {
+			i++;
+			hadmeta++;
+		}
+# endif
 		else if (quote);
 		else if (s[i] == '"') quote = s[i];
 # if	!MSDOS || !defined (_NOORIGSHELL)
@@ -1085,18 +1227,13 @@ int cx, len, plen, max, linemax, comline, cont;
 		else if (s[i] == '=' || strchr(CMDLINE_DELIM, s[i]))
 			top = i + 1;
 	}
-	if (top == cx) {
-		putterm(t_bell);
-		return(0);
-	}
 	if (comline && top > 0) {
 		for (i = top - 1; i >= 0; i--)
 			if (s[i] != ' ' && s[i] != '\t') break;
 		if (i >= 0 && !strchr(SHELL_OPERAND, s[i])) comline = 0;
 	}
 
-	cp1 = malloc2(cx - top + 1);
-	strncpy2(cp1, s + top, cx - top);
+	cp1 = strdupcpy(&(s[top]), cx - top);
 	cp1 = evalpath(cp1, 1);
 	hasmeta = 0;
 	for (i = 0; cp1[i]; i++) {
@@ -1117,15 +1254,20 @@ int cx, len, plen, max, linemax, comline, cont;
 		argc = 0;
 		l = strlen(cp1);
 		if (comline && !strdelim(cp1, 1)) {
-#ifdef	_NOORIGSHELL
+# ifdef	_NOORIGSHELL
 			argc = completeuserfunc(cp1, l, argc, &argv);
 			argc = completealias(cp1, l, argc, &argv);
 			argc = completebuiltin(cp1, l, argc, &argv);
 			argc = completeinternal(cp1, l, argc, &argv);
-#else
+# else
 			argc = completeshellcomm(cp1, l, argc, &argv);
-#endif
+# endif
 		}
+# ifndef	_NOARCHIVE
+		if (archivefile && !comline && *cp1 != _SC_)
+			argc = completearch(cp1, l, argc, &argv);
+		else
+# endif
 		argc = completepath(cp1, l, argc, &argv, comline);
 		if (!argc && comline)
 			argc = completepath(cp1, l, argc, &argv, 0);
@@ -1144,7 +1286,7 @@ int cx, len, plen, max, linemax, comline, cont;
 	cp1 = findcommon(argc, argv);
 	fix = '\0';
 	if (argc == 1 && cp1)
-		fix = ((cp2 = strrdelim(cp1, 0)) && !*(cp2 + 1)) ? _SC_ : ' ';
+		fix = ((cp2 = strrdelim(cp1, 0)) && !cp2[1]) ? _SC_ : ' ';
 
 	if (!cp1 || ((ins = (int)strlen(cp1) - ins) <= 0 && fix != ' ')) {
 		if (cont <= 0) {
@@ -1153,7 +1295,7 @@ int cx, len, plen, max, linemax, comline, cont;
 		}
 		else {
 			selectfile(argc, argv);
-			if (lcmdline > 0) {
+			if (lcmdline < 0) {
 				dispprompt(NULL, 0);
 				displaystr(s, cx, len, plen, max, linemax);
 			}
@@ -1178,8 +1320,23 @@ int cx, len, plen, max, linemax, comline, cont;
 	}
 
 	if (hasmeta) {
+		char *tmp, *home;
+		int hlen;
+
+		if (quote || quoted > top || s[top] != '~') {
+			home = NULL;
+			i = hlen = 0;
+		}
+		else {
+			tmp = &(s[top]);
+			s[len] = '\0';
+			home = malloc2(len - top + 1);
+			hlen = evalhome(&home, 0, &tmp);
+			i = ++tmp - &(s[top]);
+		}
+
 		if (quote);
-		else if (quoted) {
+		else if (quoted > top) {
 			quote = s[quoted];
 			setcursor(vlen(s, quoted), plen, max, linemax);
 			deletechar(s, quoted, len, plen, max, linemax, 1);
@@ -1187,28 +1344,54 @@ int cx, len, plen, max, linemax, comline, cont;
 			l--;
 			setcursor(vlen(s, --cx), plen, max, linemax);
 		}
-		else if (len < max) {
+		else if (len + hlen - i < max) {
+			if (home) {
+				setcursor(vlen(s, top), plen, max, linemax);
+				deletechar(s, top, len, plen, max, linemax, i);
+				delshift(s, top, len, i);
+				len -= i;
+				l -= i;
+				cx -= i;
+				i = insertstr(s, top, len, plen, max, linemax,
+					home, hlen, '\0');
+				len += i;
+				l += i;
+				cx += i;
+				free(home);
+			}
+# if	!MSDOS || !defined (_NOORIGSHELL)
+			if (hadmeta) for (i = top; i < cx; i++) {
+				if (iskanji1(s, i)) {
+					i++;
+					continue;
+				}
+				if (!isnmeta(s, i, '\0', cx)) continue;
+				if (strchr(DQ_METACHAR, s[i + 1])) continue;
+				setcursor(vlen(s, i), plen, max, linemax);
+				deletechar(s, i, len, plen, max, linemax, 1);
+				delshift(s, i, len--, 1);
+				l--;
+				cx--;
+			}
+# endif	/* !MSDOS || !_NOORIGSHELL */
 			setcursor(vlen(s, top), plen, max, linemax);
 			insertchar(s, top, len, plen, max, linemax, 1);
 			insshift(s, top, len++, 1);
 			l++;
 			s[top] = quote = '"';
-			putch2(quote);
+			putcursor(quote, 1);
 			setcursor(vlen(s, ++cx), plen, max, linemax);
 		}
 		else hasmeta = 0;
 	}
 
 	cp2 = cp1 + (int)strlen(cp1) - ins;
-	if (fix == _SC_) {
-		ins--;
-		if (!hasmeta) quote = '\0';
-	}
+	if (fix == _SC_) ins--;
 	i = insertstr(s, cx, len, plen, max, linemax, cp2, ins, quote);
 	l += i;
 	if (fix && (len += i) < max) {
 		cx += i;
-		if (quote && len + 1 < max) {
+		if (quote && len + 1 < max && (fix != _SC_ || hasmeta)) {
 			insertchar(s, cx, len, plen, max, linemax, 1);
 			insshift(s, cx, len++, 1);
 			l++;
@@ -1233,6 +1416,9 @@ char *s;
 int *cxp, cx2, *lenp, plen, max, linemax, *histnop, h;
 char **tmp;
 {
+#ifndef	_NOORIGSHELL
+	int len;
+#endif
 	int i, j;
 
 	keyflush();
@@ -1258,9 +1444,27 @@ char **tmp;
 			s[j++] = history[h][*histnop][i];
 		}
 		s[j] = '\0';
+#ifndef	_NOORIGSHELL
+		len = *lenp;
+#endif
 		*lenp = strlen(s);
 		*cxp = *lenp;
 		displaystr(s, *cxp, *lenp, plen, max, linemax);
+#ifndef	_NOORIGSHELL
+		if (shellmode) {
+			int y1, y2;
+
+			y1 = (len + plen) / linemax;
+			y2 = (*lenp + plen) / linemax;
+			if (y1 > y2) {
+				while (y1 > y2) {
+					Xlocate(0, y1--);
+					putterm(l_clear);
+				}
+				setcursor(*cxp, plen, max, linemax);
+			}
+		}
+#endif
 		cx2 = vlen(s, *cxp);
 		(*histnop)++;
 	}
@@ -1316,9 +1520,27 @@ char **tmp;
 			free(*tmp);
 			*tmp = NULL;
 		}
+#ifndef	_NOORIGSHELL
+		len = *lenp;
+#endif
 		*lenp = strlen(s);
 		*cxp = *lenp;
 		displaystr(s, *cxp, *lenp, plen, max, linemax);
+#ifndef	_NOORIGSHELL
+		if (shellmode) {
+			int y1, y2;
+
+			y1 = (len + plen) / linemax;
+			y2 = (*lenp + plen) / linemax;
+			if (y1 > y2) {
+				while (y1 > y2) {
+					Xlocate(0, y1--);
+					putterm(l_clear);
+				}
+				setcursor(*cxp, plen, max, linemax);
+			}
+		}
+#endif
 		cx2 = vlen(s, *cxp);
 	}
 	else {
@@ -1376,6 +1598,11 @@ int *cxp, cx2, *lenp, plen, max, linemax;
 {
 	int i, quote;
 
+	if (!curfilename) {
+		putterm(t_bell);
+		return(cx2);
+	}
+
 	quote = 0;
 	keyflush();
 	if (curfilename[i = 0] != '~') for (; curfilename[i]; i++) {
@@ -1388,8 +1615,8 @@ int *cxp, cx2, *lenp, plen, max, linemax;
 		s[(*cxp)++] = quote = '"';
 		putcursor(quote, 1);
 	}
-	i = insertstr(s, *cxp, *lenp, plen, max, linemax, curfilename,
-		strlen(curfilename), quote);
+	i = insertstr(s, *cxp, *lenp, plen, max, linemax,
+		curfilename, strlen(curfilename), quote);
 	if (!i) {
 		putterm(t_bell);
 		return(cx2);
@@ -1429,7 +1656,7 @@ int *cxp, cx2, len, plen, max, linemax, upper;
 		if (ch != s[*cxp]) {
 			s[(*cxp)++] = ch;
 			cx2++;
-			putch2(ch);
+			putcursor(ch, 1);
 			if (*cxp < max && (cx2 + plen) % linemax < 1)
 				setcursor(cx2, plen, max, linemax);
 			else win_x++;
@@ -1535,16 +1762,23 @@ char *s;
 int plen, max, linemax, def, comline, h;
 {
 	char *tmphist;
-	int len, cx, cx2, ocx2, i, hist, ch, ch2, quote, dupwin_x, dupwin_y;
-	int maxlen;
+	int len, cx, cx2, ocx2, i, hist, ch, ch2, quote, sig, maxlen;
 #if	!MSDOS
 	char *cp;
 	int l;
 #endif
 
-	dupwin_x = win_x;
-	dupwin_y = win_y;
 	subwindow = 1;
+
+#ifndef	_NOORIGSHELL
+	if (shellmode) sig = 0;
+	else
+#endif
+#ifdef	SIGALRM
+	if (clockmode) sig = SIGALRM;
+	else
+#endif
+	sig = 0;
 #ifndef	_NOEDITMODE
 	Xgetkey(-1, 0);
 #endif
@@ -1573,17 +1807,20 @@ int plen, max, linemax, def, comline, h;
 		if (!quote) {
 #ifndef	_NOORIGSHELL
 			if (shellmode && !maxlen) {
-				if ((ch = Xgetkey(SIGALRM, 1)) < 0) {
+				if ((ch = Xgetkey(sig, 1)) < 0) {
 					ch = K_ESC;
 					break;
 				}
 			}
 			else
 #endif
-			ch = Xgetkey(SIGALRM, 0);
+			ch = Xgetkey(sig, 0);
+#ifndef	_NOORIGSHELL
+			if (shellmode && !comline && ch == cc_intr) break;
+#endif
 		}
 		else {
-			i = ch = getkey2(SIGALRM);
+			i = ch = getkey2(sig);
 			quote = 0;
 #if	MSDOS
 			switch (i) {
@@ -1624,7 +1861,7 @@ int plen, max, linemax, def, comline, h;
 					setcursor(cx2, plen, max, linemax);
 				continue;
 			}
-#if	!MSDOS && defined (_NOKANJICONV)
+#if	!MSDOS && !defined (_NOKANJICONV)
 			else if (inputkcode == EUC
 			&& i == 0x8e && kbhit2(WAITMETA * 1000L));
 			else if (inputkcode != EUC && iskna(i));
@@ -1670,7 +1907,7 @@ int plen, max, linemax, def, comline, h;
 #ifndef	_NOCOMPLETE
 				if (selectlist) {
 					i = tmpfilepos;
-					tmpfilepos += FILEPERLOW;
+					tmpfilepos += FILEPERROW;
 					selectfile(i, NULL);
 					ocx2 = -1;
 				}
@@ -1685,7 +1922,7 @@ int plen, max, linemax, def, comline, h;
 #ifndef	_NOCOMPLETE
 				if (selectlist) {
 					i = tmpfilepos;
-					tmpfilepos -= FILEPERLOW;
+					tmpfilepos -= FILEPERROW;
 					selectfile(i, NULL);
 					ocx2 = -1;
 				}
@@ -1722,25 +1959,31 @@ int plen, max, linemax, def, comline, h;
 				break;
 			case K_DL:
 				keyflush();
-				if (cx >= len) putterm(t_bell);
-				else {
+				if (cx < len)
 					truncline(cx2, vlen(s, len), plen,
 						max, linemax);
-					len = cx;
-				}
+				maxlen = len = cx;
 				break;
-			case CTRL('L'):
-				if (dupn_line >= 0 && dupn_line != n_line)
-					ypos += n_line - dupn_line;
-				dupn_line = n_line;
+			case K_CTRL('L'):
 				keyflush();
-				Xlocate(plen, 0);
-				putterm(l_clear);
-				for (i = 0; i < WCMDLINE; i++) {
-					if (ypos + i >= n_line) break;
-					locate(xpos, ypos + i);
+#ifndef	_NOORIGSHELL
+				if (shellmode) {
+					Xlocate(0, 0);
 					putterm(l_clear);
 				}
+				else
+#endif
+				{
+					if (dupn_line >= 0
+					&& dupn_line != n_line)
+						ypos += n_line - dupn_line;
+					dupn_line = n_line;
+					for (i = 0; i < WCMDLINE; i++) {
+						if (ypos + i >= n_line) break;
+						locate(xpos, ypos + i);
+						putterm(l_clear);
+					}
+				}
 				dispprompt(NULL, 0);
 				displaystr(s, cx, len, plen, max, linemax);
 				break;
@@ -1839,16 +2082,20 @@ int plen, max, linemax, def, comline, h;
 		}
 	} while (ch != K_CR);
 
+#ifndef	_NOCOMPLETE
 	if (selectlist) selectfile(-1, NULL);
+#endif
 	setcursor(vlen(s, len), plen, max, linemax);
-	win_x = dupwin_x;
-	win_y = dupwin_y;
 	subwindow = 0;
 #ifndef	_NOEDITMODE
 	Xgetkey(-1, 0);
 #endif
 	if (tmphist) free(tmphist);
 
+#ifndef	_NOORIGSHELL
+	if (shellmode && ch == cc_intr) len = 0;
+	else
+#endif
 	if (ch == K_ESC) len = 0;
 	s[len] = '\0';
 
@@ -1862,9 +2109,11 @@ int plen;
 {
 	int i, ov;
 
-	if (!lcmdline) ov = LCMDLINE + WCMDLINE - n_line;
-	else if (lcmdline > 0) ov = lcmdline - 1 + WCMDLINE - n_line;
-	else ov = lcmdline + WCMDLINE;
+#ifndef	_NOORIGSHELL
+	if (shellmode) return(MAXLINESTR);
+#endif
+	if (lcmdline < 0) return(MAXLINESTR);
+	ov = ((lcmdline) ? lcmdline : L_CMDLINE) + WCMDLINE - n_line;
 	if (plen < 0) {
 		plen = evalprompt(NULL, promptstr, MAXLINESTR);
 #ifdef	_NOORIGSHELL
@@ -1886,12 +2135,22 @@ int set;
 	int len;
 
 	if (set) prompt = s;
+#ifndef	_NOORIGSHELL
+	if (shellmode) Xlocate(0, 0);
+	else
+#endif
 	locate(xpos, ypos);
 	if (prompt && *prompt) {
-		putch2(' ');
-		putterm(t_standout);
-		len = kanjiputs(prompt) + 1;
-		putterm(end_standout);
+#ifndef	_NOORIGSHELL
+		if (shellmode) len = kanjiputs(prompt);
+		else
+#endif
+		{
+			putch2(' ');
+			putterm(t_standout);
+			len = kanjiputs(prompt) + 1;
+			putterm(end_standout);
+		}
 	}
 	else {
 #ifdef	_NOORIGSHELL
@@ -1906,6 +2165,7 @@ int set;
 #endif
 		putterm(t_normal);
 	}
+	win_x += len;
 	return(len);
 }
 
@@ -1916,17 +2176,31 @@ char *def;
 int h;
 {
 	char *dupl, input[MAXLINESTR + 1];
-	int i, j, len, ch, comline;
+	int i, j, len, ch, comline, dupwin_x, dupwin_y;
 
-	xpos = 0;
-	if (!lcmdline) ypos = LCMDLINE;
-	else if (lcmdline > 0) ypos = lcmdline - 1;
-	else ypos = lcmdline + n_line;
+	dupwin_x = win_x;
+	dupwin_y = win_y;
+	win_x = xpos = 0;
 	dupn_line = n_line;
-	for (i = 0; i < WCMDLINE; i++) {
-		if (ypos + i >= n_line) break;
-		locate(0, ypos + i);
-		putterm(l_clear);
+#ifndef	_NOORIGSHELL
+	if (shellmode) lcmdline = -1;
+#endif
+
+	if (!lcmdline) ypos = L_CMDLINE;
+	else if (lcmdline > 0) ypos = lcmdline;
+	else ypos = n_line - 1;
+
+#ifndef	_NOORIGSHELL
+	if (shellmode) win_y = 0;
+	else
+#endif
+	{
+		win_y = ypos;
+		for (i = 0; i < WCMDLINE; i++) {
+			if (ypos + i >= n_line) break;
+			locate(0, ypos + i);
+			putterm(l_clear);
+		}
 	}
 	len = dispprompt(prompt, 1);
 	tflush();
@@ -1987,10 +2261,21 @@ int h;
 #endif
 	ch = _inputstr(input, len, cmdlinelen(len),
 		n_column - 1, ptr, comline, h);
-	if (ch != K_ESC && (!prompt || !*prompt)) cputs2("\r\n");
+	win_x = dupwin_x;
+	win_y = dupwin_y;
+
 #ifndef	_NOORIGSHELL
-	else if (shellmode);
+	if (shellmode) prompt = NULL;
 #endif
+	if (!prompt || !*prompt) {
+		if (ch != K_ESC) {
+			if (ch == cc_intr) {
+				cputs2("^C");
+				ch = K_ESC;
+			}
+			cputs2("\r\n");
+		}
+	}
 	else {
 		for (i = 0; i < WCMDLINE; i++) {
 			if (ypos + i >= n_line) break;
@@ -1999,9 +2284,9 @@ int h;
 		}
 		locate(win_x, win_y);
 		tflush();
+		if ((!lcmdline && ypos < L_CMDLINE)
+		|| (lcmdline > 0 && ypos < lcmdline + n_line)) rewritefile(1);
 	}
-	if ((!lcmdline && ypos < LCMDLINE)
-	|| (lcmdline < 0 && ypos < lcmdline + n_line)) rewritefile(1);
 	lcmdline = 0;
 
 	if (ch == K_ESC) return(NULL);
@@ -2036,14 +2321,14 @@ char *s;
 	else if (isekana(cp1, len - 1)) len--;
 #endif
 	else if (onkanji1(cp1, vlen(cp1, len) - 1)) len--;
-	strcpy(cp1 + len, cp2);
+	strcpy(&(cp1[len]), cp2);
 	return(s);
 }
 
 static VOID NEAR yesnomes(mes)
 char *mes;
 {
-	locate(0, LMESLINE);
+	locate(0, L_MESLINE);
 	putterm(l_clear);
 	putterm(t_standout);
 	kanjiputs2(mes, n_lastcolumn, -1);
@@ -2099,14 +2384,14 @@ char *fmt;
 	Xgetkey(-1, 0);
 #endif
 
-	win_y = LMESLINE;
+	win_y = L_MESLINE;
 	ret = 1;
 	do {
 		keyflush();
 		win_x = len + 1 + (1 - ret) * 2;
 		locate(win_x, win_y);
 		tflush();
-		switch (ch = Xgetkey(SIGALRM, 0)) {
+		switch (ch = Xgetkey(1, 0)) {
 			case 'y':
 			case 'Y':
 				ret = 1;
@@ -2125,7 +2410,7 @@ char *fmt;
 			case K_LEFT:
 				ret = 1;
 				break;
-			case CTRL('L'):
+			case K_CTRL('L'):
 				yesnomes(buf);
 				break;
 			default:
@@ -2133,7 +2418,7 @@ char *fmt;
 		}
 	} while (ch != K_CR);
 
-	locate(0, LMESLINE);
+	locate(0, L_MESLINE);
 	putterm(l_clear);
 	locate(win_x, win_y);
 	tflush();
@@ -2178,23 +2463,27 @@ char *s;
 	}
 	putterm(t_bell);
 
-	locate(0, LMESLINE);
+	locate(0, L_MESLINE);
 	putterm(l_clear);
 	putterm(t_standout);
 	win_x = kanjiputs(s);
-	win_y = LMESLINE;
+	win_y = L_MESLINE;
 	putterm(end_standout);
 	tflush();
 
 	keyflush();
-	getkey2(SIGALRM);
+#ifdef	SIGALRM
+	getkey2((clockmode) ? SIGALRM : 0);
+#else
+	getkey2(0);
+#endif
 	win_x = dupwin_x;
 	win_y = dupwin_y;
 	subwindow = 0;
 
 	if (no) rewritefile(1);
 	else {
-		locate(0, LMESLINE);
+		locate(0, L_MESLINE);
 		putterm(l_clear);
 		locate(win_x, win_y);
 		tflush();
@@ -2211,12 +2500,12 @@ int val[], xx[];
 {
 	int i, new;
 
-	locate(x, LMESLINE);
+	locate(x, L_MESLINE);
 	putterm(l_clear);
 	new = (num < 0) ? -1 - num : -1;
 	for (i = 0; i < max; i++) {
 		if (!str[i]) continue;
-		locate(x + xx[i] + 1, LMESLINE);
+		locate(x + xx[i] + 1, L_MESLINE);
 		if (num < 0) putch2((val[i]) ? '*' : ' ');
 		if (i != new && val[i] != num) kanjiputs(str[i]);
 		else {
@@ -2260,7 +2549,7 @@ int val[];
 	new = (num) ? *num : -1;
 	if ((new = selectmes(new, max, x, str, val, xx)) < 0) new = 0;
 
-	win_y = LMESLINE;
+	win_y = L_MESLINE;
 	do {
 		win_x = x + xx[new + 1];
 		locate(win_x, win_y);
@@ -2268,7 +2557,7 @@ int val[];
 		old = new;
 
 		keyflush();
-		switch (ch = Xgetkey(SIGALRM, 0)) {
+		switch (ch = Xgetkey(1, 0)) {
 			case K_RIGHT:
 				for (new++; new != old; new++) {
 					if (new >= max) new = 0;
@@ -2281,7 +2570,7 @@ int val[];
 					if (str[new]) break;
 				}
 				break;
-			case CTRL('L'):
+			case K_CTRL('L'):
 				if (num) selectmes(val[new], max, x,
 					str, val, xx);
 				else selectmes(-1 - new, max, x, str, val, xx);
@@ -2289,7 +2578,7 @@ int val[];
 			case ' ':
 				if (num) break;
 				val[new] = (val[new]) ? 0 : 1;
-				locate(x + xx[new] + 1, LMESLINE);
+				locate(x + xx[new] + 1, L_MESLINE);
 				putch2((val[new]) ? '*' : ' ');
 				break;
 			default:
@@ -2304,18 +2593,18 @@ int val[];
 					break;
 				}
 				val[new] = (val[new]) ? 0 : 1;
-				locate(x + xx[new] + 1, LMESLINE);
+				locate(x + xx[new] + 1, L_MESLINE);
 				putch2((val[new]) ? '*' : ' ');
 				break;
 		}
 		if (new != old) {
 			i = x + 1;
 			if (!num) i++;
-			locate(i + xx[new], LMESLINE);
+			locate(i + xx[new], L_MESLINE);
 			putterm(t_standout);
 			kanjiputs(str[new]);
 			putterm(end_standout);
-			locate(i + xx[old], LMESLINE);
+			locate(i + xx[old], L_MESLINE);
 			if (stable_standout) putterm(end_standout);
 			else kanjiputs(str[old]);
 		}
@@ -2329,7 +2618,7 @@ int val[];
 #endif
 
 	if (stable_standout) {
-		locate(x + 1, LMESLINE);
+		locate(x + 1, L_MESLINE);
 		putterm(l_clear);
 	}
 	if (num) {
@@ -2337,7 +2626,7 @@ int val[];
 		else *num = val[new];
 		for (i = 0; i < max; i++) {
 			if (!str[i]) continue;
-			locate(x + xx[i] + 1, LMESLINE);
+			locate(x + xx[i] + 1, L_MESLINE);
 			if (i == new) kanjiputs(str[i]);
 			else cprintf2("%*s", (int)strlen(str[i]), " ");
 		}
@@ -2350,7 +2639,7 @@ int val[];
 		}
 		for (i = 0; i < max; i++) {
 			if (!str[i]) continue;
-			locate(x + xx[i] + 1, LMESLINE);
+			locate(x + xx[i] + 1, L_MESLINE);
 			if (val[i]) kanjiputs(str[i]);
 			else cprintf2("%*s", (int)strlen(str[i]), " ");
 		}
