@@ -10,11 +10,11 @@
 #include <fcntl.h>
 #include "machine.h"
 
-#ifndef NOUNISTDH
+#ifndef	NOUNISTDH
 #include <unistd.h>
 #endif
 
-#ifndef NOSTDLIBH
+#ifndef	NOSTDLIBH
 #include <stdlib.h>
 #endif
 
@@ -112,6 +112,9 @@ int outputkcode = 0;
 #if	!MSDOS && !defined (_NOKANJIFCONV)
 int fnamekcode = 0;
 #endif
+#if	(!MSDOS && !defined (_NOKANJICONV)) || !defined (_NODOSDRIVE)
+char *unitblpath = NULL;
+#endif
 #ifndef	LSI_C
 u_char kctypetable[256] = {
 	0200, 0200, 0200, 0200, 0200, 0200, 0200, 0200,	/* 0x00 */
@@ -148,8 +151,8 @@ u_char kctypetable[256] = {
 	0013, 0013, 0013, 0013, 0013, 0010, 0010,    0
 };
 #endif	/* !LSI_C */
+
 #if	(!MSDOS && !defined (_NOKANJICONV)) || !defined (_NODOSDRIVE)
-char *unitblpath = NULL;
 static u_char *unitblbuf = NULL;
 static u_short unitblent = 0;
 static kconv_t rsjistable[] = {
@@ -579,7 +582,7 @@ VOID readunitable(VOID_A)
 	if (!unitblpath || !*unitblpath) strcpy(path, UNICODETBL);
 	else strcatdelim2(path, unitblpath, UNICODETBL);
 
-	if ((fd = open(path, O_RDONLY | O_BINARY, 0600)) < 0) return;
+	if ((fd = open(path, O_BINARY | O_RDONLY, 0600)) < 0) return;
 	if (read(fd, buf, 2) != 2) {
 		close(fd);
 		return;
@@ -708,7 +711,7 @@ int encode;
 		if (!unitblpath || !*unitblpath) strcpy(path, UNICODETBL);
 		else strcatdelim2(path, unitblpath, UNICODETBL);
 
-		if ((fd = open(path, O_RDONLY | O_BINARY, 0600)) < 0)
+		if ((fd = open(path, O_BINARY | O_RDONLY, 0600)) < 0)
 			return(r);
 		if (!unitblent) {
 			if (read(fd, buf, 2) != 2) {
