@@ -10,14 +10,12 @@
 #include "kctype.h"
 #include "kanji.h"
 
-#ifndef	_NOORIGSHELL
-#include "system.h"
+#if	MSDOS && !defined (_NOUSELFN)
+extern char *shortname __P_((char *, char *));
 #endif
 
-#if	MSDOS
-extern char *shortname __P_((char *, char *));
-#else
-#include <sys/param.h>
+#ifndef	_NOORIGSHELL
+#include "system.h"
 #endif
 
 extern int mark;
@@ -38,12 +36,10 @@ static int NEAR setarg __P_((char *, int, char *, char *, u_char));
 static int NEAR setargs __P_((char *, int, int, int, macrostat *));
 static int NEAR insertarg __P_((char *, char *, char *, int));
 static char *NEAR addoption __P_((char *, char *, macrostat *, int));
-#ifdef	_NOORIGSHELL
 static int NEAR system3 __P_((char *, int, int));
+#ifdef	_NOORIGSHELL
 static char *NEAR evalargs __P_((char *, int, char *[]));
 static char *NEAR evalalias __P_((char *));
-#else
-#define	system3(c, n, i)	system2(c, n)
 #endif
 
 #ifdef	_NOORIGSHELL
@@ -70,7 +66,7 @@ int ptr;
 char *arg;
 {
 	if (!arg || !*arg);
-	else if (*arg == _SC_);
+	else if (*arg == _SC_ || isdotdir(arg));
 #if	MSDOS
 	else if (_dospath(arg));
 #endif
@@ -93,6 +89,13 @@ int ptr, eol, code;
 	char *cp, tmp[MAXCOMMSTR + 1];
 	int len, total;
 
+# ifndef	_NOKANJIFCONV
+	if (code < 0) {
+		cp = _evalpath(&(buf[ptr]), &(buf[eol]), 0, 0);
+		code = getkcode(cp);
+		free(cp);
+	}
+# endif
 	buf[eol] = '\0';
 	cp = &(buf[ptr]);
 	total = 0;
@@ -110,7 +113,7 @@ int ptr, eol, code;
 	memcpy(&(buf[ptr]), tmp, total);
 	return(total);
 }
-#endif
+#endif	/* !MSDOS && !_NOKANJICONV */
 
 static int NEAR setarg(buf, ptr, dir, arg, flags)
 char *buf;
@@ -120,6 +123,9 @@ u_char flags;
 {
 #if	!MSDOS && !defined (_NOKANJIFCONV)
 	char conv[MAXPATHLEN];
+#endif
+#ifndef	_NOROCKRIDGE
+	char rbuf[MAXPATHLEN];
 #endif
 	char *cp, path[MAXPATHLEN];
 	int len, optr;
@@ -129,15 +135,13 @@ u_char flags;
 	if (!arg || !*arg) return(ptr - optr);
 	if (!dir || !*dir) {
 #if	MSDOS && !defined (_NOUSELFN)
-		if ((flags & F_TOSFN) && shortname(arg, path))
-			arg = strdup2(path);
-		else
+		if ((flags & F_TOSFN) && shortname(arg, path) == path)
+			arg = path;
 #endif
-		arg = killmeta(arg);
 	}
 	else {
 		strcatdelim2(path, dir, arg);
-		arg = killmeta(path);
+		arg = path;
 #if	MSDOS && !defined (_NOARCHIVE)
 		if (flags & F_ISARCH) for (len = 0; arg[len]; len++) {
 			if (iskanji1(arg, len)) len++;
@@ -146,13 +150,15 @@ u_char flags;
 #endif
 	}
 
+	cp = arg;
 #if	!MSDOS && !defined (_NOKANJIFCONV)
-	cp = kanjiconv2(conv, arg, MAXPATHLEN - 1, DEFCODE, getkcode(arg));
-	if (cp != arg) {
-		free(arg);
-		arg = strdup2(cp);
-	}
+	arg = kanjiconv2(conv, arg, MAXPATHLEN - 1, DEFCODE, getkcode(arg));
 #endif
+#ifndef	_NOROCKRIDGE
+	if (cp != path) arg = detransfile(cp, rbuf, 0);
+#endif
+	arg = killmeta(arg);
+
 	if ((flags & F_NOEXT) && (cp = strrchr(arg, '.')) && cp != arg)
 		len = cp - arg;
 	else len = strlen(arg);
@@ -186,7 +192,10 @@ macrostat *stp;
 	tmp[len] = '\0';
 	cp = tmp + blen;
 	flags = *(++cp);
-	for (s = ++cp; *s; s++) if (strchr(IFS_SET, *s)) break;
+	for (s = ++cp; *s; s++) {
+		if (strchr(CMDLINE_DELIM, *s)) break;
+		if (iskanji1(s, 0)) s++;
+	}
 	flen = s - cp;
 	rlen = len - (s - tmp);
 
@@ -236,16 +245,29 @@ static int NEAR insertarg(buf, format, arg, needmark)
 char *buf, *format, *arg;
 int needmark;
 {
+#if	!MSDOS && !defined (_NOKANJIFCONV)
+	char conv[MAXPATHLEN];
+#endif
 #if	MSDOS && !defined (_NOUSELFN)
-	char tmp[MAXPATHLEN];
+	char *org, alias[MAXPATHLEN];
+#endif
+#ifndef	_NOROCKRIDGE
+	char rbuf[MAXPATHLEN];
 #endif
 	char *cp, *src, *ins;
 	int i, len, ptr;
 
-	arg = killmeta(arg);
 #if	MSDOS && !defined (_NOUSELFN)
-	*tmp = '\0';
+	org = arg;
+	*alias = '\0';
 #endif
+#if	!MSDOS && !defined (_NOKANJIFCONV)
+	arg = kanjiconv2(conv, arg, MAXPATHLEN - 1, DEFCODE, getkcode(arg));
+#endif
+#ifndef	_NOROCKRIDGE
+	arg = detransfile(arg, rbuf, 0);
+#endif
+	arg = killmeta(arg);
 	ptr = strlen(format);
 	if (ptr > MAXCOMMSTR) i = 0;
 	else {
@@ -255,9 +277,14 @@ int needmark;
 			ins = arg;
 #if	MSDOS && !defined (_NOUSELFN)
 			if (*src & F_TOSFN) {
-				if (!*tmp && !shortname(arg, tmp))
-					strcpy(tmp, arg);
-				ins = tmp;
+				if (!*alias) {
+					if (shortname(org, alias) == alias)
+						org = alias;
+					org = killmeta(org);
+					strcpy(alias, org);
+					free(org);
+				}
+				ins = alias;
 			}
 #endif
 			if ((*src & F_NOEXT) && (cp = strrchr(ins, '.')))
@@ -405,14 +432,17 @@ int ignorelist;
 				else if (c == 'E') tmpcode = EUC;
 # if	FD < 2
 				else if (c == 'J') tmpcode = JIS7;
-# else
+# else	/* FD >= 2 */
 				else if (c == '7') tmpcode = JIS7;
 				else if (c == '8') tmpcode = JIS8;
 				else if (c == 'J') tmpcode = JUNET;
 				else if (c == 'H') tmpcode = HEX;
 				else if (c == 'C') tmpcode = CAP;
 				else if (c == 'U') tmpcode = UTF8;
-# endif
+#  ifndef	_NOKANJIFCONV
+				else if (c == 'A') tmpcode = -1;
+#  endif
+# endif	/* FD >= 2 */
 				else {
 					line[j++] = command[i];
 					break;
@@ -430,7 +460,7 @@ int ignorelist;
 				cnvcode = tmpcode;
 				cnvptr = j;
 				break;
-#endif
+#endif	/* !MSDOS && !_NOKANJICONV */
 			default:
 				line[j++] = command[i];
 				break;
@@ -440,7 +470,11 @@ int ignorelist;
 		j += len;
 	}
 	if (command[i]) {
-		warning(E2BIG, command);
+		if (isttyiomode) warning(E2BIG, command);
+		else {
+			errno = E2BIG;
+			perror(command);
+		}
 		return(NULL);
 	}
 
@@ -449,9 +483,13 @@ int ignorelist;
 #endif
 
 	if (needburst) for (i = c = 0; i < j; i++) {
-		if (line[i] && strchr(IFS_SET, line[i])) c = -1;
+		if (line[i] && strchr(CMDLINE_DELIM, line[i])) c = -1;
 		else if (c < 0) c = i;
-		if (line[i] || !(line[++i] & F_BURST)) continue;
+		if (line[i]) {
+			if (iskanji1(line, i)) i++;
+			continue;
+		}
+		else if (!(line[++i] & F_BURST)) continue;
 
 		len = setargs(line, c, i - 1 - c, j, stp);
 		if (len >= MAXCOMMSTR) {
@@ -466,21 +504,27 @@ int ignorelist;
 	}
 
 	if (!(flags & F_ARGSET) && arg && *arg) {
+#ifndef	_NOROCKRIDGE
+		char rbuf[MAXPATHLEN];
+#endif
 #if	!MSDOS && !defined (_NOKANJIFCONV)
 		char conv[MAXPATHLEN];
 
 		arg = kanjiconv2(conv, arg, MAXPATHLEN - 1,
 			DEFCODE, getkcode(arg));
+#endif
+#ifndef	_NOROCKRIDGE
+		arg = detransfile(arg, rbuf, 0);
 #endif
 		line[j++] = ' ';
-		if (checksc(NULL, 0, arg) < 0) arg = killmeta(arg);
-		else {
+		arg = killmeta(arg);
+		if (checksc(NULL, 0, arg) >= 0) {
 			cp = malloc2(strlen(arg) + 2 + 1);
 			cp[0] = '.';
 			cp[1] = _SC_;
 			strcpy(&(cp[2]), arg);
-			arg = killmeta(cp);
-			free(cp);
+			free(arg);
+			arg = cp;
 		}
 
 		len = strlen(arg);
@@ -503,27 +547,22 @@ int ignorelist;
 	return(cp);
 }
 
-char *inputshellstr(prompt, ptr, cursor, def)
+char *inputshellstr(prompt, ptr, def)
 char *prompt;
-int ptr, cursor;
+int ptr;
 char *def;
 {
 	char *cp, *tmp, *duppromptstr;
-	int x, y;
 
 	duppromptstr = promptstr;
-	if (prompt) {
-		promptstr = prompt;
-		lcmdline = n_line;
-	}
-	if (cursor && getxy(&y, &x) >= 0) lcmdline = y;
+	if (prompt) promptstr = prompt;
 	cp = inputstr(NULL, 0, ptr, def, 0);
 	promptstr = duppromptstr;
 	if (!cp) return((char *)-1);
 
-	stdiomode();
+	stdiomode(1);
 	tmp = evalhistory(cp);
-	ttyiomode();
+	ttyiomode(1);
 	if (!tmp) {
 		free(cp);
 		return(NULL);
@@ -543,7 +582,7 @@ char *def;
 #endif
 	char *cp;
 
-	cp = inputshellstr(NULL, ptr, 0, def);
+	cp = inputshellstr(NULL, ptr, def);
 	if (cp == (char *)-1) return(NULL);
 	else if (!cp) {
 		hideclock = 1;
@@ -561,12 +600,12 @@ char *def;
 		if (cp) {
 			trp = analyze(cp, trp, 1);
 			free(cp);
-			if (!trp || !(trp -> flags & ST_CONT)) break;
+			if (!trp || !(trp -> cont)) break;
 		}
 
-		lcmdline = n_line;
+		lcmdline = -1;
 		hideclock = 1;
-		cp = inputshellstr(promptstr2, -1, 0, NULL);
+		cp = inputshellstr(promptstr2, -1, NULL);
 		if (cp == (char *)-1) break;
 		else if (!cp) continue;
 
@@ -629,7 +668,9 @@ int ignorelist;
 		}
 
 		if (p < 0) p = strlen(line);
+		if (!(i = isttyiomode)) ttyiomode(1);
 		cp = inputstr("", 0, p, line, 0);
+		if (!i) stdiomode(1);
 		if (!cp) return(NULL);
 		if (!*cp) {
 			free(cp);
@@ -646,11 +687,11 @@ int ignorelist;
 	return(command);
 }
 
-#ifdef	_NOORIGSHELL
 static int NEAR system3(command, noconf, ignorelist)
 char *command;
 int noconf, ignorelist;
 {
+#ifdef	_NOORIGSHELL
 	char *cp, *tmp;
 	int n, ret;
 
@@ -666,8 +707,46 @@ int noconf, ignorelist;
 	sigvecset(n);
 	if (cp) free(cp);
 	return(ret);
-}
+#else	/* _NOORIGSHELL */
+	if (!isttyiomode) return(dosystem(command));
+	return(system2(command, noconf));
 #endif	/* _NOORIGSHELL */
+}
+
+int isinternalcomm(command)
+char *command;
+{
+#ifndef	_NOORIGSHELL
+	syntaxtree *stree;
+	char **subst;
+	int i, *len;
+#endif
+	char **argv;
+	int ret;
+
+	ret = 1;
+#ifdef	_NOORIGSHELL
+	if (getargs(command, &argv) > 0 && checkinternal(argv[0]) < 0) ret = 0;
+	freevar(argv);
+#else	/* !_NOORIGSHELL */
+	stree = newstree(NULL);
+	if (analyze(command, stree, 1)) {
+		if ((argv = getsimpleargv(stree))) {
+			i = (stree -> comm) -> argc;
+			(stree -> comm) -> argc =
+				getsubst(i, argv, &subst, &len);
+			stripquote(argv[0], 1);
+			if (argv[0] && checkinternal(argv[0]) < 0) ret = 0;
+			freevar(subst);
+			free(len);
+		}
+	}
+	freestree(stree);
+	free(stree);
+#endif	/* !_NOORIGSHELL */
+
+	return(ret);
+}
 
 int execmacro(command, arg, noconf, argset, ignorelist)
 char *command, *arg;
@@ -675,7 +754,7 @@ int noconf, argset, ignorelist;
 {
 	static char *duparg = NULL;
 	macrostat st;
-	char *tmp, **argv, buf[MAXCOMMSTR + 1];
+	char *tmp, buf[MAXCOMMSTR + 1];
 	int i, max, r, ret, status;
 
 	if (arg) duparg = arg;
@@ -688,34 +767,7 @@ int noconf, argset, ignorelist;
 	}
 	n_args = mark;
 
-	st.flags = F_ARGSET;
-	if (!argset) {
-#ifdef	_NOORIGSHELL
-		if (getargs(command, &argv) > 0 && checkinternal(argv[0]) < 0)
-			st.flags = 0;
-		freevar(argv);
-#else	/* !_NOORIGSHELL */
-		syntaxtree *stree;
-		char **subst;
-		int *len;
-
-		stree = newstree(NULL);
-		if (analyze(command, stree, 1)) {
-			if ((argv = getsimpleargv(stree))) {
-				i = (stree -> comm) -> argc;
-				(stree -> comm) -> argc =
-					getsubst(i, argv, &subst, &len);
-				stripquote(argv[0], 1);
-				if (argv[0] && checkinternal(argv[0]) < 0)
-					st.flags = 0;
-				freevar(subst);
-				free(len);
-			}
-		}
-		freestree(stree);
-		free(stree);
-#endif	/* !_NOORIGSHELL */
-	}
+	st.flags = (argset || isinternalcomm(command)) ? F_ARGSET : 0;
 	if (noconf < 0) st.flags |= F_ISARCH;
 
 	if (!(tmp = evalcommand(command, duparg, &st, ignorelist))) {
@@ -724,7 +776,7 @@ int noconf, argset, ignorelist;
 	}
 	if (noconf >= 0 && (st.flags & F_NOCONFIRM)) noconf = 1 - noconf;
 	st.flags |= F_ARGSET;
-	if (noconf >= 0 && !argset) {
+	if (noconf >= 0 && argset <= 0) {
 		if (!(tmp = addoption(tmp, duparg, &st, ignorelist))) {
 			if (arg) duparg = NULL;
 			return(-1);
@@ -743,7 +795,7 @@ int noconf, argset, ignorelist;
 		if (!(st.flags & F_REMAIN)
 		|| !(tmp = evalcommand(command, duparg, &st, ignorelist)))
 			break;
-		if (noconf >= 0 && !argset) {
+		if (noconf >= 0 && argset <= 0) {
 			if (!(tmp = addoption(tmp, duparg, &st, ignorelist)))
 				break;
 		}
@@ -787,7 +839,8 @@ char *argv[];
 	while ((cp = strtkchr(&(command[i]), '$', 1))) {
 		while (&(command[i]) < cp && j < MAXCOMMSTR)
 			line[j++] = command[i++];
-		if (command[++i] < '0' || command[i] > '9') {
+		i++;
+		if (!isdigit(command[i])) {
 			if (j > MAXCOMMSTR - 2) break;
 			line[j++] = '$';
 			line[j++] = command[i++];
@@ -981,10 +1034,7 @@ int *ptrp;
 	ptr = (ptrp) ? *ptrp : 0;
 	size = histsize[0];
 	if (str[ptr] == '!') n = 0;
-	else if (str[ptr] == '=' || strchr(IFS_SET, str[ptr])) {
-		ptr = -1;
-		n = -1;
-	}
+	else if (str[ptr] == '=' || strchr(IFS_SET, str[ptr])) n = ptr = -1;
 	else if ((cp = evalnumeric(&(str[ptr]), &no, -1))) {
 		if (!no) n = -1;
 		else if (no < 0) {
@@ -1001,7 +1051,7 @@ int *ptrp;
 			if (strchr(CMDLINE_DELIM, str[ptr + i])) break;
 			if (iskanji1(str, ptr + i)) i++;
 		}
-		if (!i) ptr = -1;
+		if (!i) n = ptr = -1;
 		else {
 			for (n = 0; n < size; n++) {
 				if (!history[0][n]) break;
