@@ -29,8 +29,8 @@
 #define	STDERR_FILENO	2
 #endif
 
-#define	CR	'\r'
-#define	ESC	'\033'
+#define	K_CR	'\r'
+#define	K_ESC	'\033'
 
 #define	K_MIN	K_NOKEY
 #define	K_NOKEY	0401
@@ -118,8 +118,6 @@ extern char *l_insert;
 extern char *l_delete;
 extern char *c_insert;
 extern char *c_delete;
-extern char *c_store;
-extern char *c_restore;
 extern char *c_home;
 extern char *c_locate;
 extern char *c_return;
@@ -149,6 +147,7 @@ extern FILE *ttyout;
 #define	putterm(s)	cputs2(s)
 #define	putterms(s)	cputs2(s)
 #else
+extern int tputs __P_((char *, int, int (*)__P_((int))));
 #define	putterm(s)	tputs(s, 1, putch3)
 #define	putterms(s)	tputs(s, n_line, putch3)
 #endif
@@ -200,7 +199,7 @@ extern int chgcolor __P_((int, int));
 #define	WAITKEYPAD	360		/* msec */
 #endif
 #ifndef	WAITMETA
-#define	WAITMETA	60		/* msec */
+#define	WAITMETA	120		/* msec */
 #endif
 
 #define	ANSI_BLACK	0
