@@ -22,7 +22,6 @@
 #include <dpmi.h>
 #include <go32.h>
 #include <sys/farptr.h>
-#define	FAR
 #define	NOP	0x00
 #define	RETF	0xc3
 # if	(DJGPP >= 2)
@@ -40,7 +39,7 @@
 #define	PTR_FAR(ptr)		((u_long)(__tb))
 #define	PTR_SEG(ptr)		(__tb_segment)
 #define	PTR_OFF(ptr, ofs)	(__tb_offset + (ofs))
-#else
+#else	/* !DJGPP */
 # ifdef	__TURBOC__	/* Oops!! Borland C++ has not x.bp !! */
 typedef union DPMI_REGS {
 	struct XREGS {
@@ -54,13 +53,12 @@ typedef union DPMI_REGS {
 typedef union REGS	__dpmi_regs;
 # endif
 #define	__attribute__(x)
-#define	FAR	far
 #define	NOP	0x90
 #define	RETF	0xcb
 #define	PTR_FAR(ptr)		(((u_long)FP_SEG(ptr) << 4) + FP_OFF(ptr))
 #define	PTR_SEG(ptr)		FP_SEG(ptr)
 #define	PTR_OFF(ptr, ofs)	FP_OFF(ptr)
-#endif
+#endif	/* !DJGPP */
 
 #define	FR_CARRY	00001
 #define	FR_PARITY	00004
@@ -181,7 +179,7 @@ typedef struct _partition_t {
 	u_char t_sect[4] __attribute__ ((packed));
 #endif
 } partition_t;
-#define	PART_SIZE	sizeof(partition_t)
+#define	PART_SIZE	((int)sizeof(partition_t))
 
 typedef struct _xparam_t {
 	u_char size[2] __attribute__ ((packed));
@@ -202,7 +200,7 @@ typedef struct _xparam_t {
 	u_char reserved2 __attribute__ ((packed));
 	u_char checksum __attribute__ ((packed));
 } xparam_t;
-#define	XPARAM_SIZE	sizeof(xparam_t)
+#define	XPARAM_SIZE	((int)sizeof(xparam_t))
 
 typedef struct _xpacket_t {
 	u_char size __attribute__ ((packed));
@@ -211,7 +209,7 @@ typedef struct _xpacket_t {
 	u_char bufptr[4] __attribute__ ((packed));
 	u_char sect[8] __attribute__ ((packed));
 } xpacket_t;
-#define	XPACKET_SIZE	sizeof(xpacket_t)
+#define	XPACKET_SIZE	((int)sizeof(xpacket_t))
 
 typedef struct _drvinfo {
 	u_long head;
